unit secd_cod;
interface

const
  s=0;  e=1;  c=2;  { SEC itself     } 
 w1=3;  a=4;  b=5;  { work and cache }
 r0=6; r1=7; io=8;  { running lists, io queue }
 

 Debug:boolean=false;
 ShdMaxCnt=127;

var
 LoopCnt, SwapCnt,
 CPCnt, 
 PendingCnt, PendMax:longint;
 SwapOK:boolean;

procedure RTError(n:word; msg:string);
procedure secd_loop;
procedure ShowSECD;

{ Cache memory vars & functions: }

const ChCnt:byte=0;
procedure Just(n:byte);

{ 
  ------------ SECD commands summary: -------------------

  Subcommands:
  
  CleanX:
    repeat
      ([nil.x] s) e c --> (x s) e c
    until S is ([nil.x] s);
    if S is ([p.q] s) then begin
      q --> ((S e c) q)    (dalay current process until recipe [p.q] is not ready) 
      Load pending process from r1;
    end;

 
  Commands:

  RTN (#rtn) --> executed if C=nil
  
    normal:=true;
    while normal and S is (x [p.q] s) do begin
      CleanX;
      if S is ([c.e] [p1.q1] s)
        then normal:=false
        else begin
          if x is a node
             then replace [p.q] with x
             else replace [p.q] with [nil.x];
          (x y s) e nil --> (x s) e nil;
          while q<>nil do roll(q,r1);
          if p>0 then Primitive(p);   
        end;
    end; 
    if normal
       then (x e' c' s) e nil  --> (x s) e' c'
       else ([c.e] s) e' nil   --> ([0.nil] s) e c



  1 LD x y       s  e c           --> ((GetNth(GetNth(e,x),y) s) e c  
  2 LDC          s  e (x c)       --> (x s)     e    c 
  3 LDF          s  e (f c)       --> (f.e s)   e    c 
  4 DELAY        s  e (c c')      --> ([c.e] s) e    c' 
  5 CASE_k       if i in [0..k-1] then c_i:=ci else c_i:=ck
                 (i s) e (c0 ... ck c) --> (c_i s)  e c
  6 DUM          s  e c           --> s      (nil e) c 
  7 STOP         Kill the process. Attempt to load some pending process. 
  8 CALL        
         if c=nil
          then   (c' s) e nil     --> s       e c'
          else   (c' s) e c       --> (e c s) e c'

  9 RECV  (Pipe s)   e c     --> Receive message from Pipe 
          (Stream s) e c     --> Receive lexeme from Stream 
          (Connwait s) e c   --> Receive tcp connection from listener or connection try 
 10 SEND  (Pipe msg s) e c   --> Send message to Pipe
          (Stream lex s) e c --> Send lexeme to Stream
 11 RECVC (Stream s) e c     --> Receive char (ascii code) from Stream 
 12 SENDC (Stream I s) e c   --> Send char with ascii code I to Stream
 13 FORK  s e (c1 c2 c)      --> (c1 s) e c
                                 r0 --> (((c2 s) e c) r0)
 14 POP   (x s) e c          --> s e c


 ---------------------------- Binary primitives -----------------------------
  
 16 CONS    (x y s) e c   --> (cons(x,y) s) e c  
 17 EQ      (x y s) e c   --> (eq(x,y)   s) e c
 18 FOPEN   (m fn s) e c  --> ([stream:f.nil] s) e c  --> open file fn with mode m
 19 LISTEN  (P M s) e c   --> ([connwait:socket.nil] s) e c  
            Start tcp listener on port P, allowing maximum M connections
 20 CONNECT (P H s) e c   --> ([connwait:socket.nil] s) e c  
            Try tcp connection to host with ip address H on port P
 21
 22 ADD     (x y s) e c   --> (y+x s)       e c 
 23 SUB     (x y s) e c   --> (y-x s)       e c 
 24 MUL     (x y s) e c   --> (y*x s)       e c 
 25 DIV     (x y s) e c   --> ((y div x) s) e c 
 26 REM     (x y s) e c   --> ((y mod x) s) e c 
 27 LEQ     (x y s) e c   --> ((y <= x) s)  e c  --> for integers 
 28 LEXLEQ  (x y s) e c   --> ((y <= x) s)  e c  --> for atom names
 29 
 30 CONCAT  (a b s) e c   --> (a||b s)      e c  --> concatenates atom names
 31
 
 ---------------------------- APP / TR_APP codes: ---------------------------

 32 LET n        s         e (c' c)    --> (e c Drop(s,n))       (s e)       c' 
 33 TRLET n      s         e c         --> Drop(s,n)             (s e)       c  
 34 LETREC n     s         e (c' c)    --> ((cdr e) c Drop(s,n)) Setcar(e,s) c' 
 35 TRLETR n     s         e c         --> Drop(s,n)             Setcar(e,s) c  
 36 APP n        (c'.e' s) e c         --> (e c Drop(s,n))       (s e')      c' 
 37 TR_APP n     (c'.e' s) e c         --> Drop(s,n)             (s e')      c' 
 38  
 39  
 40 RECOVER (#rec_p - p is unary primitive. Now only #rec_0 is used ! )
    CleanX;
    if S is ([c.e] s)
       then ([c.e] s) e' c' --> ([p.nil] e' c' s) e c
       else (x s) e c       --> (p(x) s) e c   
 
 41 TRREC   (#trrec_p - p is primitive or 
             128+n if encode (... #rec #trapp_n) sequence )
    CleanX;
    if S is ([c.e] s)
       then ([c.e] s) e' c' --> ([p.nil] s) e c
       else begin
         (x s) e c       --> (p(x) s') e c  
         if p<128 then RTN
       end; 

 42 
 43

 ---------------------------- Unary primitives ------------------------------

 48 CAR     (x s) e c  --> (car(x) s)   e c
 49 CDR     (x s) e c  --> (cdr(x) s)   e c
 50 INT     (x s) e c  --> (int?(x) s)  e c
 51 ATOM    (x s) e c  --> (atom?(x) s) e c
 52 NULL    (x s) e c  --> ((x=nil?) s) e c   
 53 NODE    (x s) e c  --> (node?(x) s) e c
 54 FCLOSE  (f s) e c  --> (close(f) s) e c   close file/tcp stream
 55 STAT    (x s) e c  --> (x s)        e c   for debuging
 56 INC     (x s) e c  --> (x+1 s)      e c 
 57 DEC     (x s) e c  --> (x-1 s)      e c    
 58 CHR     (n s) e c  --> (chr(n) s)   e c   chr(n) is a one char atom
 59 EOF     (f s) e c  --> (eof?(f) s)  e c   checks for eof / broken connection       
 60   
 61
 62
 63
}

implementation

uses secd_mem, secd_str, secd_msg, secd_tcp;

procedure ShowSECD;
 const CMax=80;
 procedure ShowCache;
  begin
   case ChCnt of
    0:begin
       write(stderr,'S=  '); 
      end;
    1:begin
       write(stderr,'Sa= '); ShowReg(a,CMax,false); writeln(stderr);
       write(stderr,'St= '); 
      end;
    2:begin
       write(stderr,'Sa= '); ShowReg(a,CMax,false); writeln(stderr);
       write(stderr,'Sb= '); ShowReg(b,CMax,false); writeln(stderr);
       write(stderr,'St= '); 
      end;
   end; 
   ShowReg(s,CMax,false); writeln(stderr);
  end;
 begin
  writeln(stderr,'SECD: Loops=',LoopCnt,'  Cons=',ConsCnt,
               '   Runs=',PendingCnt); 
  ShowCache;
  write(stderr,'E=  '); ShowReg(e,CMax,false); writeln(stderr);
  write(stderr,'C=  '); ShowReg(c,CMax,false); writeln(stderr);
  {readln;} 
 end;

procedure Just(n:byte);
 begin if ChCnt<>n then begin
  case ChCnt of
   0:if n=1 
      then Pop(s,a)
      else begin Pop(s,a); Pop(s,b) end;
   1:if n=0 
      then Push(a,s) else Pop(s,b); 
   2:if n=1 
      then Push(b,s)
      else begin Push(b,s); Push(a,s) end;
  end;
  ChCnt:=n;
 end end;

procedure TRApp(par:word); 
 begin { TRAPP:   (c'.e' s) e c --> Drop(s,n) (s e')    c' }
  Just(1);
  MoveReg(a,e); 
  Pop(e,c);
  Push(s,e); 
  Drop(s,par);
  ChCnt:=0;
 end; 

procedure RTError(n:word; msg:string);
 begin
  writeln(stderr);
  writeln(stderr,msg,' !');
  ShowSECD;
  Flush(stderr);
         
  Just(0);
  SetAtom(a,null,0);
  Push(c,a);
  Push(e,a);
  Push(s,a);
  SetAtom(b,int,n);
  Push(b,a);
  SetAtom(s,null,0);
  Push(a,s); { S' = ((n S E C)) }
  while Regs[e].t=node do Pop(e,b);
  while Regs[b].t=node do Pop(b,a); { A = err_handler }
  ChCnt:=1;
  TRApp(1);

  secd_loop;
 end;

procedure secd_loop;
 var
  cmd:word;
  par:longint;

 procedure Primitive(cmd:byte);
  procedure SetBool(b:boolean);
   begin
    if b then SetAtom(a,int,0) else SetAtom(a,int,1);
   end;
  procedure UniCmd;
   begin
    if ChCnt=0 then Just(1);
    if (cmd in [56..58,60]) and (Regs[a].t<>int)
     then RTError(0,'The argument must be int');
    if (cmd in [54,59]) and 
       (not ((Regs[a].t=node) and (SubTag(a)=stream)))
     then RTError(45,'The argument must be a stream');
    case cmd of
      48:CarSelf(a);
      49:Drop(a,1);
      50:SetBool(Regs[a].t=int);
      51:SetBool(Regs[a].t=atom);
      52:SetBool(Regs[a].t=null);
      53:SetBool(Regs[a].t=node);
      54:begin
          Car(a,w1);
          FClose(PStream(Regs[w1].v));
          SetAtom(w1,int,longint(nil));
          SetCar(a,w1);
         end;
      55:ShowStat;   
      56:inc(Regs[a].v);
      57:dec(Regs[a].v);
      58:NewAtom(chr(Regs[a].v),a);
      59:begin
          Car(a,w1);
          SetBool(FEof(PStream(Regs[w1].v)));
         end;
    end;     
   end;
  procedure BinCmd;
   var ws:string; 
   begin
    Just(2);
    if cmd in [19..20,22..27] then begin
     if Regs[a].t<>int then RTError(1,'Second argument must be int'); 
     if Regs[b].t<>int then RTError(2,'First argument must be int'); 
    end;
    if cmd in [28,30] then begin
     if Regs[a].t<>atom then RTError(3,'Second argument must be atom'); 
     if Regs[b].t<>atom then RTError(4,'First argument must be atom'); 
    end;
    case cmd of
     16:PushR(b,a);
     17:SetBool((Regs[b].t=Regs[a].t) and (Regs[b].v=Regs[a].v)); 
     18:begin
         if Regs[b].t<>int then 
          RTError(5,'First argument must be TStreamType: 0-input 1-output'); 
         case Regs[a].t of
          null:ws:='';
          atom:ws:=GetStr(Regs[a].v);
         end;
         CreateIO(longint(FOpen(ws,TStreamType(Regs[b].v))),stream);
        end; 
     19:CreateIO(longint(TCPListen( Regs[a].v,Regs[b].v)),connwait);
     20:CreateIO(longint(TCPConnect(Regs[a].v,Regs[b].v)),connwait);
     22:Regs[a].v:=Regs[b].v+Regs[a].v;
     23:Regs[a].v:=Regs[b].v-Regs[a].v;
     24:Regs[a].v:=Regs[b].v*Regs[a].v;
     25:Regs[a].v:=Regs[b].v div Regs[a].v;
     26:Regs[a].v:=Regs[b].v mod Regs[a].v;
     27:SetBool(Regs[b].v <= Regs[a].v);
     28:SetBool(GetStr(Regs[b].v) <= GetStr(Regs[a].v));
     30:NewAtom(GetStr(Regs[b].v)+GetStr(Regs[a].v),a);
    end;     
    ChCnt:=1;
   end;
  begin
   if cmd<128 
    then case cmd of
      16..31:BinCmd;
      48..63:UniCmd;
    end
    else TRApp(cmd-128);   
  end;
 procedure CleanX;
  var finish:boolean; 
  begin
   repeat
    if TestRec(a) then begin
      Car(a,w1);
      if Regs[w1].t=null 
       then begin Drop(a,1); finish:=false end
       else begin
        finish:=true;
        if Regs[w1].t=int then begin
         Cdr(a,w1);
         Push(e,c);
         Just(0);
         Push(s,c);
         Car(s,a);
         Push(c,w1); 
         SetCdr(a,w1); { [p.q] --> [p.((s e c) q)] } 
         SwapOK:=true;
        end;
       end; 
    end else finish:=true;
   until finish;
  end;
 procedure Rec;
  begin
   if TestRec(a) then begin
    Just(1);
    MoveReg(e,b); 
    Push(c,s);
    StartRec(par);
    ChCnt:=2;
   end else if par>0 then Primitive(par);
  end; 
 procedure TR_Rec;
  begin
   if TestRec(a) then StartRec(par)
   else if par>0 then Primitive(par);
  end;
 procedure Rtn;
  var p:byte; normalRTN:boolean;
  begin
   normalRTN:=true;
   Just(2);
   while normalRTN and TestRec(b) do begin
    CleanX;
    if SwapOK then Exit;
    if TestRec(a) and TestRec(b)
     then normalRTN:=false
     else begin
      p:=ReplaceRec;
      while Regs[w1].t=node do begin
       roll(w1,r1); IncPend;
      end; 
      { Specialy for nodes the replaced recipe is returned: }
      if Regs[a].t=node then MoveReg(b,a);
      Pop(s,b);
      if p>0 then begin 
        if p>127 then RTError(6,'#trapp_n into #rtn'); 
        Primitive(p); Just(2) 
      end;  
     end;
   end;
   if normalRTN then begin
    MoveReg(b,e);
    Pop(s,c);
    ChCnt:=1;
   end else StartRec(0);
  end;
 procedure LDx;
  begin
   if ChCnt=2 then Just(1);
   inc(ChCnt);
   if ChCnt=2 then MoveReg(a,b); 
   case cmd of
    1:begin
       GetNth(e,w1,par shr 8);
       GetNth(w1,a,par and $ff);
      end;
    2:Pop(c,a);
    3:begin Pop(c,a); PushR(e,a) end;
    4:begin Pop(c,a); PushR(e,a); SetSubTag(a,recipe) end; 
   end;
  end; 
 procedure CASE_k;
  var k:longint;
  begin
   if ChCnt=0 then Just(1);
   if Regs[a].t=int then k:=Regs[a].v else k:=par;
   if (k<0) or (k>par) then k:=par;
   GetNth(c,a,k);
   Drop(c,par+1);
  end;
 procedure CALL; 
  begin
   if Regs[c].t=node then begin { normal CALL }
    Just(1);
    MoveReg(c,b); 
    MoveReg(a,c);
    CopyReg(e,a);
    ChCnt:=2;
   end else begin               { TR CALL }   
    if ChCnt=0 then Just(1);
    MoveReg(a,c); dec(ChCnt);
    if ChCnt=1 then MoveReg(b,a); 
   end;
  end;
 procedure FORK;
  begin
   Just(0);
   Pop(c,a); 
   CopyReg(s,w1);
   Roll(c,s); 
   CopyReg(c,b);
   ProcSave;
   MoveReg(b,c);
   MoveReg(w1,s);
   Inc(ChCnt);
  end;
 procedure SpecCmd;
  begin
   case cmd of
     1..4:LDx;
     5:CASE_k; 
     6:begin SetAtom(w1,null,-1); Push(w1,e) end;
     7:; { stop - Phase0 }
     8:CALL;
     9:RecvMsg(0);
    10:SendMsg(0);
    11:RecvMsg(1);
    12:SendMsg(1);
    13:FORK;
    14:begin { POP }
        if ChCnt=0 then Just(1);
        dec(ChCnt);
        if ChCnt=1 then MoveReg(b,a); 
       end;
   end;
  end;
 procedure CallCmd;
  begin
   case cmd of
    32:begin { LET:     s e (c' c)  --> (e c Drop(s,n)) (s e) c'  }
        Just(0);
        Cdr(c,b);
        CopyReg(e,a); 
        Push(s,e);
        Drop(s,par);
        CarSelf(c);
        ChCnt:=2;
       end;
    33:begin { TRLET:   s e c       --> Drop(s,n) (s e) c       }
        Just(0); 
        Push(s,e); 
        Drop(s,par);
       end;
    34:begin { LETREC:  s e (c' c)  --> ((cdr e) c Drop(s,n)) Setcar(e,s) c' }
        Just(0); 
        Cdr(c,b);
        Cdr(e,a); 
        SetCar(e,s);
        Drop(s,par);
        CarSelf(c);
        ChCnt:=2;
       end;
    35:begin { TRLETR:  s e c       --> Drop(s,n) Setcar(e,s) c  } 
        Just(0); 
        SetCar(e,s); 
        Drop(s,par);
       end;
    36:begin { APP:     (c'.e' s) e c --> (e c Drop(s,n)) (s e') c' }
        Just(1);
        MoveReg(c,b); 
        MoveReg(a,c); 
        MoveReg(e,a); 
        Cdr(c,e); 
        Push(s,e);
        CarSelf(c);
        Drop(s,par);
        ChCnt:=2;
       end; 
    37:TRApp(par); { TRAPP:   (c'.e' s) e c --> Drop(s,n) (s e') c' }
    38:;
    39:;
    40:Rec;
    41:TR_Rec;
    42:;
    43:;
   end;
  end;
 procedure Phase0;
  begin
   case cmd of
     7:SwapOK:=true;
    40,41:begin { Rec / TR_Rec }
       if ChCnt=0 then Just(1);
       CleanX;
      end;
   end;
  end;
 procedure Phase1;
  begin
   Drop(c,1); 
   case cmd of
     1..15:SpecCmd;
    16..31:Primitive(cmd);
    32..47:CallCmd;
    48..63:Primitive(cmd);
   end;
  end;
 begin
  SwapOK:=false;
{  
  PendingCnt:=1;
  PendMax:=PendingCnt;
  CPCnt:=0;
}  
  repeat
   if Debug then ShowSECD; 
   inc(LoopCnt);
   inc(CPCnt);

   { Test if timelimit is reached and if there is another 
     running or sleaping processes: }  
   if (CPCnt>ShdMaxCnt) and 
      ((Regs[r0].t=node) or
       (Regs[r1].t=node) or
       (Regs[io].t=node)) 
    then begin
     inc(SwapCnt);
     Just(0);
     ProcSave;
     IO_Scan;
     SwapOK:=true;
     if debug then writeln(stderr,'----------- SWAP -------------');
    end
    else if Regs[c].t=null 
     then Rtn
     else begin
      Car(c,w1);
      if Regs[w1].t<>int 
       then RTError(7,'Command not found at the top of C') 
       else begin
        cmd:=Regs[w1].v shr 24;
        par:=Regs[w1].v and $ffffff;
        Phase0;
        if not SwapOK then Phase1;
       end;
     end;

   if SwapOK then ProcLoad;
   
  until SwapOK;
  Just(0);
 end;

begin
 SwapCnt:=0;
 LoopCnt:=0;
 PendingCnt:=1;
 PendMax:=PendingCnt;
 CPCnt:=0;
end.


