unit sec_drivers;

interface

uses Sockets, Linux, sec_atoms;

type
 PTCPSocket = ^TTCPSocket;
 TCPConnTyp  = (listener, conn_from, conn_to);
 TTCPSocket = record
  conn:longint;
  SA:TInetSockAddr;
  t:TCPConnTyp;
  pcnt:word;
  broken:boolean;
 end;

function TCPListen (aPort:word; maxconn:word):PTCPSocket;
function TCPConnect(aPort:word; host:longint):PTCPSocket;
function TCPGetConn(socket:PTCPSocket; var ok:boolean):PTCPSocket;

type
 PStream = ^TStream;
 TStreamType = (instream, outstream);
 TFileDescrT = (stdio, regfile, tcpconn);
 TStream = record
  t:TStreamType;
  sp:PTCPSocket;
  fd:longint;
  ef:boolean;
  fdt:TFileDescrT;
  lex:TAtomS;
  w:char;
  srmode:byte;
 end;

function  FOpen(fn:string;         st:TStreamType):PStream;
function  SOpen(socket:PTCPSocket; st:TStreamType):PStream;
procedure FClose(p:PStream);
function  FGetLex (p:PStream; var ok:boolean):string;
function  FGetChar(p:PStream; var ok:boolean):char;
function  FPutStr(p:PStream; s:string):boolean;
function  FPutChar(p:PStream; c:char):boolean;
function  FEof(p:PStream):boolean;
function  StreamCont(ps:PStream):boolean;

procedure Delay(n:word);

implementation

uses sec_lists, sec_codes;

{ ------------- TCP drivers: ------------- }

const
 Err_TryAgain=11;
 Err_InProgress=115;
 
function SwapWord(w:word):word;
 begin
  SwapWord:=(w shl 8) or (w shr 8);
 end;
    
function TCPSocket(aPort:word):PTCPSocket;
 var p:PTCPSocket;
 begin
  New(p);
  with p^ do begin
   conn:=Socket(AF_INET,SOCK_STREAM,0 {6} {PF_INET});
   if conn<0 then RTError(14,'socket');
   fcntl(conn,F_SetFl,Open_NonBlock);
   with SA do begin
    family:=AF_INET;
    port:=SwapWord(aPort);
   end; 
  end; 
  TCPSocket:=p;
 end;
 
function TCPListen (aPort:word; maxconn:word):PTCPSocket;
 var p:PTCPSocket;
 begin
  p:=TCPSocket(aPort);
  with p^ do begin
   SA.addr:=0; { allow connections from anywhere }
   t:=listener;
   if not Bind(conn,SA,SizeOf(SA)) then RTError(15,'bind');
   if not Listen(conn,maxconn)     then RTError(16,'listen');
  end;  
  TCPListen:=p;
 end;
 
function TCPConnect(aPort:word; host:longint):PTCPSocket;
 var p:PTCPSocket;
 begin
  p:=TCPSocket(aPort);
  with p^ do begin
   SA.addr:=host; { ((1 shl 24) or 127) for localhost } 
   t:=conn_to;
  end;  
  TCPConnect:=p;
 end;
 
function TCPGetConn(socket:PTCPSocket; var ok:boolean):PTCPSocket;
 var p:PTCPSocket; addrlen:longint;
 begin 
  case socket^.t of
   listener:begin
     New(p);
     p^.t:=conn_from;
     addrlen:=SizeOf(p^.SA);
     p^.conn:=Accept(socket^.conn,p^.SA,addrlen);
     p^.pcnt:=0;
     p^.broken:=false;
     ok:=p^.conn>=0;
     if ok 
      then begin
       TCPGetConn:=p;
       writeln(stderr,'TCP fd=',p^.conn); flush(stderr);
      end 
      else if SocketError<>Err_TryAgain 
       then RTError(17,'accept')
       else Dispose(p);     
    end;
   conn_to:begin
     with socket^ do begin
      ok:=Connect(conn,SA,SizeOf(SA));
      pcnt:=0;
      broken:=false;
     end;
     if (not ok) and (SocketError<>Err_InProgress) then begin
      writeln(stderr,'SocketError=',SocketError);
      RTError(18,'connect');
     end;
     TCPGetConn:=socket;
    end;
  end;  
 end;

{ ---------- File drivers: --------- }

const SpecChar=['(',')','.',' '];

function NewStream(
          aST:TStreamType;
          aSP:PTCPSocket;
          aFD:longint;
          aFDT:TFileDescrT):PStream;
 var p:PStream;
 begin
  New(p);
  with p^ do begin
   t:=aST;
   ef:=false;
   lex:='';
   w:=' ';
   fdt:=aFDT;
   sp :=aSP;
   fd :=aFD;
   fcntl(fd,F_SetFl,Open_NonBlock);
  end;
  NewStream:=p;
 end;

function  FOpen(fn:string; st:TStreamType):PStream;
 begin case st of
  instream: if fn=''
   then FOpen:=NewStream(st,nil,0,stdio)
   else FOpen:=NewStream(st,nil,fdOpen(fn,Open_RdOnly),regfile);
  outstream: if fn=''
   then FOpen:=NewStream(st,nil,1,stdio)
   else begin
    UnLink(fn);
    FOpen:=NewStream(st,nil,fdOpen(fn,Open_WrOnly or Open_Creat),regfile);
   end; 
 end end;

function  SOpen(socket:PTCPSocket; st:TStreamType):PStream;
 begin 
  SOpen:=NewStream(st,socket,socket^.conn,tcpconn);
  inc(socket^.pcnt);
 end;

procedure CTest(p:PStream);
 begin
  if p=nil then RTError(32,'File is allready closed'); 
 end;
 
procedure FClose(p:PStream);
 begin
  CTest(p);
  with p^ do case fdt of
   stdio:;
   regfile:fdClose(fd);
   tcpconn:begin
     dec(sp^.pcnt);
     if sp^.pcnt=0 then begin
      shutdown(fd,2);
      fdClose(fd);
      Dispose(sp);
     end;
    end;
  end; 
  Dispose(p);
 end;

function  FEof(p:PStream):boolean;
 begin
  CTest(p);
  FEof:=false;
  with p^ do case fdt of
   stdio:;
   regfile:if t=instream then FEof:=ef;
   tcpconn:FEof:=sp^.broken;
  end; 
 end;

procedure SetEof(p:PStream);
 begin with p^ do case fdt of
  regfile:ef:=true;
  tcpconn:sp^.broken:=true;
 end end;

procedure fdRdErr;
 begin
  writeln(stderr,'LinuxError=',LinuxError);
  RTError(33,'fdRead')
 end;
      
function  FGetLex(p:PStream; var ok:boolean):string;
 var Ready:boolean; ow:char; 
 procedure rd;
  begin with p^ do begin
   Ready:=true;
   if FEof(p) 
    then w:=')' 
    else if fdRead(fd,w,1)<>1  
     then case LinuxError of
      0:begin SetEof(p); w:=' ' end;
      Err_TryAgain:Ready:=false;
      else fdRdErr;
     end; 
   if (w=chr(10)) or (w=chr(13)) then w:=' ';
  end end;
 begin 
  CTest(p);
  with p^ do
   if t=outstream then RTError(34,'recv from output stream')
   else begin
    srmode:=0;
    ok:=false; Ready:=true;
    while Ready and (not ok) do 
     if (lex='') and (w=' ') then rd else 
      if w in SpecChar
       then if lex=''
        then begin 
         ow:=w; rd;
         if Ready then begin 
          FGetLex:=ow; 
          ok:=true 
         end; 
        end
        else begin 
         FGetLex:=lex; 
         ok:=true; lex:='' 
        end
       else begin 
        ow:=w; rd;
        if Ready then lex:=lex+ow; 
       end
   end;
 end;

function  FGetChar(p:PStream; var ok:boolean):char;
 begin 
  CTest(p);
  with p^ do
   if t=outstream then RTError(35,'recvc from output stream')
   else begin
    srmode:=1;
    ok:=true;
    if FEof(p) 
     then w:=')' 
     else if fdRead(fd,w,1)<>1  
      then case LinuxError of
       0:begin SetEof(p); w:=' ' end;
       Err_TryAgain:ok:=false;
       else fdRdErr;
      end; 
    FGetChar:=w; 
   end;
 end; 

procedure fdWrErr;
 begin
  writeln(stderr,'LinuxError=',LinuxError);
  RTError(36,'fdWrite')
 end;
      
function FPutStr(p:PStream; s:string):boolean;
 var n,m:longint;
 begin 
  CTest(p);
  with p^ do if t=instream 
   then RTError(37,'send to input stream')
   else begin
    srmode:=0;
    n:=length(s);
    if FEof(p)
     then FPutStr:=true
     else begin
      m:=fdWrite(fd,s[1],n);
      if m=n 
       then FPutStr:=true
       else case LinuxError of
        0:SetEof(p); 
        Err_TryAgain:begin
          FPutStr:=false;
          lex:=copy(s,m+1,n-m);
         end; 
        else fdWrErr;
       end;
      end; 
     end;  
 end;

function  FPutChar(p:PStream; c:char):boolean;
 begin 
  CTest(p);
  with p^ do if t=instream 
   then RTError(39,'sendc to input stream')
   else begin
    srmode:=1;
    if FEof(p)
     then FPutChar:=true
     else if 1=fdWrite(fd,c,1)
      then FPutChar:=true
      else case LinuxError of
       0:SetEof(p);
       Err_TryAgain:begin FPutChar:=false; w:=c end; 
       else fdWrErr;
      end;
   end;  
 end;

function StreamCont(ps:PStream):boolean;
 var 
  ok:boolean; 
  ws:string;
  wc:char;
 begin 
  case ps^.t of 
   instream:case ps^.srmode of
     0:begin
        ws:=FGetLex(ps,ok);
        if ok then NewAtom(ws,w1); { w1 = GetLex }
       end;
     1:begin
        wc:=FGetChar(ps,ok);
        if ok then SetAtom(w1,int,ord(wc)); { w1 = ord(GetChar) }
       end;
    end;
   outstream:begin
     case ps^.srmode of
      0:ok:=FPutStr(ps,ps^.lex);
      1:ok:=FPutChar(ps,ps^.w);
     end;
     if ok then SetAtom(w1,int,0); { w1 = true }
    end; 
  end;
  StreamCont:=ok;
 end;

{ ---------- Time drivers: -------------- }

procedure Delay(n:word);
 begin
  Select(0,nil,nil,nil,n);
 end;

{ ---------- end of all drivers --------- }

end.
