unit secd_cod;
interface

const
  s=0;  e=1;  c=2;  { SEC itself     } 
 w1=3;  a=4;  b=5;  { work and cache }
 r0=6; r1=7; io=8;  { running lists, io queue }
 

 Debug:boolean=false;
 ShdMaxCnt=127;

var
 LoopCnt, SwapCnt,
 CPCnt, 
 PendingCnt, PendMax:longint;
 SwapOK:boolean;

procedure secd_loop;
procedure ShowSECD;

{ Cache memory vars & functions: }

const ChCnt:byte=0;
procedure Just(n:byte);

{ 
  ------------ SECD commands summary: -------------------

  Subcommands:
  
  CleanX:
    repeat
      ([nil.x] s) e c --> (x s) e c
    until S is ([nil.x] s);
    if S is ([p.q] s) then begin
      q --> ((S e c) q)    (dalay current process until recipe [p.q] is not ready) 
      Load pending process from r1;
    end;

 
  Commands:

  RTN (#rtn) --> executed if C=nil
  
    normal:=true;
    while normal and S is (x [p.q] s) do begin
      CleanX;
      if S is ([c.e] [p1.q1] s)
        then normal:=false
        else begin
          if x is a node
             then replace [p.q] with x
             else replace [p.q] with [nil.x];
          (x y s) e nil --> (x s) e nil;
          while q<>nil do roll(q,r1);
          if p>0 then Primitive(p);   
        end;
    end; 
    if normal
       then (x e' c' s) e nil  --> (x s) e' c'
       else ([c.e] s) e' nil   --> ([0.nil] s) e c



  1 LD x y       s  e c           --> ((GetNth(GetNth(e,x),y) s) e c  
  2 LDC          s  e (x c)       --> (x s)     e    c 
  3 LDF          s  e (f c)       --> (f.e s)   e    c 
  4 DELAY        s  e (c c')      --> ([c.e] s) e    c' 
  5 
  6 DUM          s           e c  --> s      (nil e) c 
  7 STOP         Kill the process. Attempt to load some pending process. 
  8 
  9 RECV  (Pipe s) e c        --> ???       see secd_msg.pp
 10 SEND  (Pipe msg s) e c --> ???       see secd_msg.pp

 ---------------------------- Binary primitives -----------------------------
  
 16 CONS    (x y s) e c   --> (cons(x,y) s) e c  
 17 EQ      (x y s) e c   --> (eq(x,y)   s) e c
 18 FOPEN   (m fn s) e c  --> ([stream:f.nil] s) e c  --> open file fn with mode m
 19 
 20
 21
 22 ADD     (x y s) e c   --> (y+x s)       e c 
 23 SUB     (x y s) e c   --> (y-x s)       e c 
 24 MUL     (x y s) e c   --> (y*x s)       e c 
 25 DIV     (x y s) e c   --> ((y div x) s) e c 
 26 REM     (x y s) e c   --> ((y mod x) s) e c 
 27 LEQ     (x y s) e c   --> ((y <= x) s)  e c  --> for integers 
 28 LEXLEQ  (x y s) e c   --> ((y <= x) s)  e c  --> for atom names
 29 
 30 CONCAT  (a b s) e c   --> (a||b s)      e c  --> concatenates atom names
 31
 
 ---------------------------- APP / TR_APP codes: ---------------------------

 32 LET n        s         e (c' c)    --> (e c Drop(s,n))       (s e)       c' 
 33 TRLET n      s         e c         --> Drop(s,n)             (s e)       c  
 34 LETREC n     s         e (c' c)    --> ((cdr e) c Drop(s,n)) Setcar(e,s) c' 
 35 TRLETR n     s         e c         --> Drop(s,n)             Setcar(e,s) c  
 36 APP n        (c'.e' s) e c         --> (e c Drop(s,n))       (s e')      c' 
 37 TR_APP n     (c'.e' s) e c         --> Drop(s,n)             (s e')      c' 
 38 SEL          (x s)     e (c1 c2 c) --> (e c s)               e           cx 
 39 TR_SEL       (x s)     e (c1 c2)   --> s                     e           cx 
 
 40 RECOVER (#rec_p - p is unary primitive. Now only #rec_0 is used ! )
    CleanX;
    if S is ([c.e] s)
       then ([c.e] s) e' c' --> ([p.nil] e' c' s) e c
       else (x s) e c       --> (p(x) s) e c   
 
 41 TRREC   (#trrec_p - p is primitive or 
             128+n if encode (... #rec #trapp_n) sequence )
    CleanX;
    if S is ([c.e] s)
       then ([c.e] s) e' c' --> ([p.nil] s) e c
       else begin
         (x s) e c       --> (p(x) s') e c  
         if p<128 then RTN
       end; 

 42 FORK          s e (c1 c2 c)  --> (e c s) e c1
                                     r0 --> (((e c s) e c2) r0)

 43 TRFORK        s e (c1 c2)    --> s e c1
                                     r0 --> ((s e c2) r0)

 ---------------------------- Unary primitives ------------------------------

 48 CAR     (x s) e c  --> (car(x) s)   e c
 49 CDR     (x s) e c  --> (cdr(x) s)   e c
 50 INT     (x s) e c  --> (int?(x) s)  e c
 51 ATOM    (x s) e c  --> (atom?(x) s) e c
 52 NULL    (x s) e c  --> ((x=nil?) s) e c   
 53 NODE    (x s) e c  --> (node?(x) s) e c
 54 FCLOSE  (f s) e c  --> (close(f) s) e c   close file stream

 56 INC     (x s) e c  --> (x+1 s)      e c 
 57 DEC     (x s) e c  --> (x-1 s)      e c    
 58 CHR     (n s) e c  --> (chr(n) s)   e c   chr(n) is a one char atom
 59       
 60   

}

implementation
uses secd_mem, secd_str, secd_msg;

procedure ShowSECD;
 const CMax=100;
 procedure ShowCache;
  begin
   case ChCnt of
    0:begin
       write(stderr,'S=  '); 
      end;
    1:begin
       write(stderr,'Sa= '); ShowReg(a,CMax,false); writeln(stderr);
       write(stderr,'St= '); 
      end;
    2:begin
       write(stderr,'Sa= '); ShowReg(a,CMax,false); writeln(stderr);
       write(stderr,'Sb= '); ShowReg(b,CMax,false); writeln(stderr);
       write(stderr,'St= '); 
      end;
   end; 
   ShowReg(s,CMax,false); writeln(stderr);
  end;
 begin
  writeln(stderr,'SECD: Loops=',LoopCnt,'  Cons=',ConsCnt,
               '   Runs=',PendingCnt); 
  ShowCache;
  write(stderr,'E=  '); ShowReg(e,CMax,false); writeln(stderr);
  write(stderr,'C=  '); ShowReg(c,CMax,false); writeln(stderr);
  {readln;} 
 end;

procedure Just(n:byte);
 begin if ChCnt<>n then begin
  case ChCnt of
   0:if n=1 
      then Pop(s,a)
      else begin Pop(s,a); Pop(s,b) end;
   1:if n=0 
      then Push(a,s) else Pop(s,b); 
   2:if n=1 
      then Push(b,s)
      else begin Push(b,s); Push(a,s) end;
  end;
  ChCnt:=n;
 end end;

procedure secd_loop;
 var
  cmd:word;
  par:longint;
  OK:boolean;
  
 procedure Test;
  var s1:string;  
  begin
   if ChCnt=0 then Just(1);
   if Regs[a].t<>atom then Error('If exp is not boolean (T/F)');
   s1:=GetStr(Regs[a].v);
   OK:=(s1='T') or (s1='t');
   dec(ChCnt);
   if ChCnt=1 then MoveReg(b,a); 
  end;
 procedure TRApp(par:word); 
  begin { TRAPP:   (c'.e' s) e c --> Drop(s,n) (s e')    c' }
   Just(1);
   MoveReg(a,e); 
   Pop(e,c);
   Push(s,e); 
   Drop(s,par);
   ChCnt:=0;
  end; 
 procedure Primitive(cmd:byte);
  procedure UniCmd;
   begin
    if ChCnt=0 then Just(1);
    if cmd in [56..58,60] then if Regs[a].t<>int 
     then Error('The argument must be int'); 
    case cmd of
     48:CarSelf(a);
     49:Drop(a,1);
     50:if  Regs[a].t=int  then NewAtom('T',a) else NewAtom('F',a);
     51:if  Regs[a].t=atom then NewAtom('T',a) else NewAtom('F',a);
     52:if  Regs[a].t=null then NewAtom('T',a) else NewAtom('F',a);
     53:if  Regs[a].t=node then NewAtom('T',a) else NewAtom('F',a);
     54:if  SubTag(a)=stream then begin
         Car(a,w1);
         FClose(PStream(Regs[w1].v));
        end;
     56:inc(Regs[a].v);
     57:dec(Regs[a].v);
     58:NewAtom(chr(Regs[a].v),a);
    end;     
   end;
  procedure BinCmd;
   var ws:string; 
   begin
    Just(2);
    if cmd in [22..27] then begin
     if Regs[a].t<>int then Error('Second argument must be int'); 
     if Regs[b].t<>int then Error( 'Furst argument must be int'); 
    end;
    if cmd in [28,30] then begin
     if Regs[a].t<>atom then Error('Second argument must be atom'); 
     if Regs[b].t<>atom then Error( 'Furst argument must be atom'); 
    end;
    case cmd of
     16:PushR(b,a);
     17:if (Regs[b].t=Regs[a].t) and (Regs[b].v=Regs[a].v) 
         then NewAtom('T',a) else NewAtom('F',a);
     18:begin
         if Regs[b].t<>int then Error( 'Furst argument must be TStreamType: 0-input 1-output '); 
         case Regs[a].t of
          null:ws:='';
          atom:ws:=GetStr(Regs[a].v);
         end;
         SetAtom(a,int,longint(FOpen(ws,TStreamType(Regs[b].v))));
         SetAtom(b,null,0);
         PushR(b,a);
         SetSubTag(a,stream);
        end; 
     22:Regs[a].v:=Regs[b].v+Regs[a].v;
     23:Regs[a].v:=Regs[b].v-Regs[a].v;
     24:Regs[a].v:=Regs[b].v*Regs[a].v;
     25:Regs[a].v:=Regs[b].v div Regs[a].v;
     26:Regs[a].v:=Regs[b].v mod Regs[a].v;
     27:if Regs[b].v <= Regs[a].v
         then NewAtom('T',a) else NewAtom('F',a);
     28:if GetStr(Regs[b].v) <= GetStr(Regs[a].v)
         then NewAtom('T',a) else NewAtom('F',a);
     30:NewAtom(GetStr(Regs[b].v)+GetStr(Regs[a].v),a);
    end;     
    ChCnt:=1;
   end;
  begin
   if cmd<128 
    then case cmd of
      16..31:BinCmd;
      48..63:UniCmd;
    end
    else TRApp(cmd-128);   
  end;
 procedure CleanX;
  var finish:boolean; 
  begin
   repeat
    if TestRec(a) then begin
      Car(a,w1);
      if Regs[w1].t=null 
       then begin Drop(a,1); finish:=false end
       else begin
        finish:=true;
        if Regs[w1].t=int then begin
         Cdr(a,w1);
         Push(e,c);
         Just(0);
         Push(s,c);
         Car(s,a);
         Push(c,w1); 
         SetCdr(a,w1); { [p.q] --> [p.((s e c) q)] } 
         SwapOK:=true;
        end;
       end; 
    end else finish:=true;
   until finish;
  end;
 procedure Rec;
  begin
   if TestRec(a) then begin
    Just(1);
    MoveReg(e,b); 
    Push(c,s);
    StartRec(par);
    ChCnt:=2;
   end else if par>0 then Primitive(par);
  end; 
 procedure TR_Rec;
  begin
   if TestRec(a) then StartRec(par)
   else if par>0 then Primitive(par);
  end;
 procedure Rtn;
  var p:byte; normalRTN:boolean;
  begin
   normalRTN:=true;
   Just(2);
   while normalRTN and TestRec(b) do begin
    CleanX;
    if SwapOK then Exit;
    if TestRec(a) and TestRec(b)
     then normalRTN:=false
     else begin
      p:=ReplaceRec;
      while Regs[w1].t=node do begin
       roll(w1,r1); IncPend;
      end; 
      Pop(s,b);
      if p>0 then begin 
        if p>127 then Error('#trapp_n into #rtn'); 
        Primitive(p); Just(2) 
      end;  
     end;
   end;
   if normalRTN then begin
    MoveReg(b,e);
    Pop(s,c);
    ChCnt:=1;
   end else StartRec(0);
  end;
 procedure LDx;
  begin
   if ChCnt=2 then Just(1);
   inc(ChCnt);
   if ChCnt=2 then MoveReg(a,b); 
   case cmd of
    1:begin
       GetNth(e,w1,par shr 8);
       GetNth(w1,a,par and $ff);
      end;
    2:Pop(c,a);
    3:begin Pop(c,a); PushR(e,a) end;
    4:begin Pop(c,a); PushR(e,a); SetSubTag(a,recipe) end; 
   end;
  end; 
 procedure SpecCmd;
  begin
   case cmd of
     1..4:LDx;
     5:; { rtn - Phase0 }
     6:begin SetAtom(w1,null,-1); Push(w1,e) end;
     7:; { stop - Phase0 }
     9:RecvMsg;
    10:SendMsg;
   end;
  end;
 procedure CallCmd;
  begin
   case cmd of
    32:begin { LET:     s e (c' c)  --> (e c Drop(s,n)) (s e) c'  }
        Just(0);
        Cdr(c,b);
        CopyReg(e,a); 
        Push(s,e);
        Drop(s,par);
        CarSelf(c);
        ChCnt:=2;
       end;
    33:begin { TRLET:   s e c       --> Drop(s,n) (s e) c       }
        Just(0); 
        Push(s,e); 
        Drop(s,par);
       end;
    34:begin { LETREC:  s e (c' c)  --> ((cdr e) c Drop(s,n)) Setcar(e,s) c' }
        Just(0); 
        Cdr(c,b);
        Cdr(e,a); 
        SetCar(e,s);
        Drop(s,par);
        CarSelf(c);
        ChCnt:=2;
       end;
    35:begin { TRLETR:  s e c       --> Drop(s,n) Setcar(e,s) c  } 
        Just(0); 
        SetCar(e,s); 
        Drop(s,par);
       end;
    36:begin { APP:     (c'.e' s) e c --> (e c Drop(s,n)) (s e') c' }
        Just(1);
        MoveReg(c,b); 
        MoveReg(a,c); 
        MoveReg(e,a); 
        Cdr(c,e); 
        Push(s,e);
        CarSelf(c);
        Drop(s,par);
        ChCnt:=2;
       end; 
    37:TRApp(par); { TRAPP:   (c'.e' s) e c --> Drop(s,n) (s e') c' }
    38:begin { SEL:     (x s) e (c1 c2 c) --> (e c s) e cx }
        Test;
        Just(0);
        if OK
         then begin MoveReg(c,b); Pop(b,c); Drop(b,1) end 
         else begin Drop(c,1);   Cdr(c,b); CarSelf(c) end;
        CopyReg(e,a);
        ChCnt:=2;
       end;
    39:begin { TRSEL:   (x s) e (c1 c2)   --> s e cx }
        Test;
        if OK then CarSelf(c) else Drop(c,1);
       end;
    40:Rec;
    41:TR_Rec;
    42:begin { FORK   s e (c1 c2 c)  --> (e c s) e c1
                                         r0 --> (((e c s) e c2) r0) }
        Just(0);
        Pop(c,a); Pop(c,b);
        Push(c,s); Push(e,s);
        MoveReg(b,c); ProcSave;
        MoveReg(a,c);
       end;                                 
    43:begin { TRFORK s e (c1 c2)    --> s e c1
                                         r0 --> ((s e c2) r0) }
        Just(0);
        Pop(c,a); ProcSave; MoveReg(a,c);
       end;                                  
   end;
  end;
 procedure Phase0;
  begin
   case cmd of
   { 5:Rtn; }
     7:SwapOK:=true;
    40,41:begin { Rec / TR_Rec }
       if ChCnt=0 then Just(1);
       CleanX;
      end;
   end;
  end;
 procedure Phase1;
  begin
   Drop(c,1); 
   case cmd of
     1..15:SpecCmd;
    16..31:Primitive(cmd);
    32..47:CallCmd;
    48..63:Primitive(cmd);
   end;
  end;
 begin
  SwapOK:=false;
  PendingCnt:=1;
  PendMax:=PendingCnt;
  CPCnt:=0;
  repeat
   if Debug then ShowSECD; 
   inc(LoopCnt);
   inc(CPCnt);

   { Test if timelimit is reached and if there is another 
     running or sleaping processes: }  
   if (CPCnt>ShdMaxCnt) and 
      ((Regs[r0].t=node) or
       (Regs[r1].t=node) or
       (Regs[io].t=node)) 
    then begin
     inc(SwapCnt);
     Just(0);
     ProcSave;
     IO_Scan;
     SwapOK:=true;
     if debug then writeln(stderr,'----------- SWAP -------------');
    end
    else if Regs[c].t=null 
     then Rtn
     else begin
      Car(c,w1);
      if Regs[w1].t<>int 
       then Error('Command not found at the top of C') 
       else begin
        cmd:=Regs[w1].v shr 24;
        par:=Regs[w1].v and $ffffff;
        Phase0;
        if not SwapOK then Phase1;
       end;
     end;

   if SwapOK then ProcLoad;
   
  until SwapOK;
  Just(0);
 end;

begin
 SwapCnt:=0;
 LoopCnt:=0;
end.


