unit secd_msg;

interface

uses secd_mem;

procedure SendMsg;
procedure RecvMsg;
procedure ProcLoad;
procedure ProcSave;
procedure IncPend;
procedure IO_Scan;

{

 Pipe is passive - (q0.q1) or waiting - [pipe: s' (e'.c')]
 
 SEND
 (Pipe   msg s) e c --> ((msg s) e c)
 (Stream msg s) e c --> ((T   s) e c)

 RECV
 (Pipe   s) e c --> ((msg_from(Pipe) s) e c)
 (Stream s) e c --> ((getlex(Stream) s) e c) 

}

implementation

uses secd_cod, secd_str;

procedure ProcLoad;
 begin
  if Regs[r1].t=null then while Regs[r0].t=node do Roll(r0,r1);
  while (Regs[r1].t=null) and (Regs[io].t=node) do begin
   IO_Scan;
   if Regs[r1].t=null then Delay;
  end;
  if Regs[r1].t=node then begin
   Pop(r1,c);
   Pop(c,s);
   Pop(c,e);
   ChCnt:=0;
   SwapOK:=false;
  end;
  if debug then begin
   writeln(stderr,'---- New Proc ---- ProcCnt=',PendingCnt);
   ShowSECD;
  end; 
  CPCnt:=0;
  dec(PendingCnt); 
 end;

procedure IncPend;
 begin
  inc(PendingCnt);
  if PendingCnt>PendMax then PendMax:=PendingCnt;
 end;
 
procedure ProcSave;
 begin
  Push(e,c);
  Push(s,c);
  Push(c,r0); 
  IncPend;
 end;

procedure IOSave;
 begin
  Push(e,c);
  Push(s,c);
  SetCdr(a,c);
  Push(a,io); 
  SwapOK:=true;
 end;

procedure SendMsg;
 var ws:string;
 begin
  Just(1);
  case SubTag(a) of
   cons,pipe:
    ProcSave; { Delay current process: r0 --> (((msg s) e c) r0) }
  end;  
  ChCnt:=0;
  Pop(s,b); { a=Pipe/Stream, b:=msg } 
  case SubTag(a) of
   cons:begin
      Car(a,w1);
      Push(b,w1);
      SetCar(a,w1);
      SwapOK:=true;
     end;
   pipe:begin
      Car(a,s);
      Cdr(a,c); Pop(c,e);
      SetAtom(w1,null,0);
      SetCar(a,w1); SetCdr(a,w1);
      SetSubTag(a,cons);
      CPCnt:=0;
      MoveReg(b,a);
      ChCnt:=1;
     end;
   stream:begin
      Car(a,w1);
      case Regs[b].t of
       null:ws:='nil';
       int:ws:=GetLong(Regs[b].v);
       atom:ws:=GetStr(Regs[b].v);
       node:Error('Sending node to stream is impossible');
      end; 
      if FPutStr(PStream(Regs[w1].v),ws) 
       then begin NewAtom('T',a); ChCnt:=1 end 
       else IOSave;
     end;  
  end; 
 end;

procedure RecvMsg;
 var ok:boolean; ws:string;
 begin
  Just(1);
  ChCnt:=0;
  case SubTag(a) of
   cons:begin
      Car(a,b); Cdr(a,w1);
      if Regs[w1].t=null then while Regs[b].t=node do Roll(b,w1);
      if Regs[w1].t=null then begin
       Push(e,c);
       SetCar(a,s); SetCdr(a,c);
       SetSubTag(a,pipe);
       SwapOK:=true;
      end else begin
       SetCar(a,b); 
       Pop(w1,b);
       SetCdr(a,w1);
       MoveReg(b,a);
       ChCnt:=1;
      end;
     end;
   pipe:Error('Somebody wait on this pipe');
   stream:begin
      Car(a,w1);
      ws:=FGetLex(PStream(Regs[w1].v),ok);
      if ok 
       then begin NewAtom(ws,a); ChCnt:=1 end 
       else IOSave;
     end;
  end; 
 end;

procedure IO_Scan;
 var ok:boolean; ws:string; ps:PStream;
 procedure IOLoad(ws:string);
  begin
   Cdr(b,c);       { c  = (s' e' c') }
   NewAtom(ws,w1); { w1 = GetLex or 'T' }
   Pop(c,s); 
   Push(w1,s);
   Push(s,c);      { c  = ((w1 s') e' c') }
   SetAtom(e,null,0);
   SetCdr(b,e);    { b= <in/outfile|pFile.nil> } 
   Push(c,r1);     { process ready }
   IncPend;
  end;
 begin
  SetAtom(a,null,0);
  while Regs[io].t=node do begin
   Pop(io,b); { b = <in/outfile|pFile.(s' e' c')> }
   Car(b,w1); { w1 = pFile }
   ps:=PStream(Regs[w1].v);
   case ps^.t of 
    instream:
      begin
       ws:=FGetLex(ps,ok);
       if ok then IOLoad(ws) else Push(b,a);
      end;
    outstream:
      if FPutStr(ps,ps^.ws) then IOLoad('T') else Push(b,a);
   end;
  end;
  MoveReg(a,io);
 end;

end.

