unit secd_str;
interface

var 
 AtomPtr: longint;

function GetStr(v:longint):string;
function GetLong(v:longint):string;
procedure NewAtom(s:string; reg:byte);

procedure ReadSFrom(fname:string; r:byte);
procedure WriteSTo (fname:string; r:byte);

type
 PStream = ^TStream;
 TStreamType = (instream, outstream);
 TStream = record
  t:TStreamType;
  fd:longint;
  ef:boolean;
  std:boolean;
  ws:string;
  w:char;
 end;

function  FOpen(fn:string; st:TStreamType):PStream;
procedure FClose(p:PStream);
function  FGetLex(p:PStream; var ok:boolean):string;
function  FPutStr(p:PStream; s:string):boolean;
procedure Delay;

implementation

{ ---------- OS dependant code --------- }

uses secd_mem, Linux;

type
 TAtomS=string[32];
const 
 SpecChar=['(',')','.',' '];
 MaxStr = 500;
var
 Atoms: array[0..MaxStr] of TAtomS;

const
 Err_TryAgain = 11;
 
function  FOpen(fn:string; st:TStreamType):PStream;
 var p:PStream;
 begin
  New(p);
  with p^ do begin
   t:=st;
   std:=fn='';
   ef:=false;
   ws:='';
   w:=' ';
   case t of 
    instream: if std then fd:=0 else fd:=fdOpen(fn,Open_RdOnly);
    outstream: if std then fd:=1 else fd:=fdOpen(fn,Open_WrOnly or Open_Creat);
   end;
   fcntl(fd,F_SetFl,Open_NonBlock);
  end;
  FOpen:=p;
 end;
 
procedure FClose(p:PStream);
 begin
  with p^ do if not std then fdClose(fd);
  dispose(p);
 end;
 
function  FGetLex(p:PStream; var ok:boolean):string;
 var Ready:boolean;  
 procedure rd;
  begin with p^ do begin
   Ready:=true;
   if ef 
    then w:=')' 
    else if fdRead(fd,w,1)<>1
     then case LinuxError of
      0:begin ef:=true; w:=')' end;
      Err_TryAgain:Ready:=false;
      else begin
       writeln(stderr,'LinuxError=',LinuxError);
       Error('fdRead')
      end;
     end; 
   if (w=chr(10)) or (w=chr(13)) then w:=' ';
  end end;
 begin with p^ do
  if t=outstream then Error('GetLex from output stream')
  else begin
   ok:=false; Ready:=true;
   while Ready and (not ok) do 
    if (ws='') and (w=' ') then rd else 
     if w in SpecChar
      then if ws=''
       then begin FGetLex:=w; ok:=true; rd end
       else begin FGetLex:=ws; ok:=true; ws:='' end
      else begin ws:=ws+w; rd end
  end;
 end;

function FPutStr(p:PStream; s:string):boolean;
 var n,m:word;
 begin with p^ do
  if t=instream 
   then Error('PutStr to input stream')
   else begin
    n:=length(s);
    m:=fdWrite(fd,s[1],length(s));
    if m=n 
     then FPutStr:=true
     else case LinuxError of
      0:Error('Remote end closed connection');
      Err_TryAgain:begin
        FPutStr:=false;
        ws:=copy(s,m+1,n-m);
       end; 
      else begin
       writeln(stderr,'LinuxError=',LinuxError);
       Error('fdWrite')
      end;
     end;
   end;  
 end;

procedure Delay;
 begin
  Select(0,nil,nil,nil,10);
 end;

{ ---------- end of OS dependant code --------- }

function GetStr(v:longint):string;
 begin GetStr:=Atoms[v] end;

const
 MnemMax=63;
 SECDmnem:array[1..MnemMax] of TAtomS=
 (
  {
    RTN code is represented by null C register !!!
  }
  { 01..15 - Special codes }
  'ld-2',
  'ldc',
  'ldf',
  'delay',
  'none',
  'dum',
  'stop',
  
  'none',
  'recv',
  'send',
  'none',
  'none',
  'none',
  'none',
  'none',
  
  { 16..31 - Binary primitives }
  'cons',
  'eq',
  'fopen',
  'none',
  'none',
  'none',
  'add',
  'sub',
  
  'mul',
  'div',
  'rem',
  'leq',
  'lexleq',
  'none',
  'concat',
  'none',

  { 32..47 - Calls }  
  'let-b',
  'trlet-b',
  'letrec-b',
  'trletr-b',
  'app-b',
  'trapp-b',
  'sel',
  'trsel',
  
  'rec-p',
  'trrec-p',
  'fork',
  'trfork',
  'none',
  'none',
  'none',
  'none',

  { 48..63 - Unary primitives }
  'car',
  'cdr',
  'int',
  'atom',
  'null',
  'node',
  'fclose',
  'none',

  'inc',
  'dec',
  'chr',
  'none',
  'none',
  'none',
  'none',
  'none'
 ); 
 MnemBase=$FFFFFF;


function EatStr(var s:string; dl:char):string;
 var n:byte;
 begin
  n:=pos(dl,s);
  if n>0 then begin
   EatStr:=copy(s,1,n-1);
   delete(s,1,n);
   while (length(s)>0) and (s[1]=dl) do delete(s,1,1);
  end else begin
   EatStr:=s;
   s:='';
  end;
 end;

function MnemHead(i:byte):string;
 var c1:string;
 begin
  c1:=SECDmnem[i];
  MnemHead:=EatStr(c1,'-');
 end; 
function MnemTail(i:byte):string;
 var c1:string;
 begin
  c1:=SECDmnem[i];
  EatStr(c1,'-');
  MnemTail:=c1;
 end; 

function GetLong(v:longint):string;
 var s,s1,mt:string; n:longint;
 begin
  if v>MnemBase then begin
   n:=v shr 24;
   if n>MnemMax then Error('Big SECD code');
   s:='#'+MnemHead(n);
   mt:=MnemTail(n);
   if mt<>'' then case mt[1] of
    'p':begin
         n:=v and MnemBase;
         if n in [1..MnemMax] 
          then s:=s+'_'+MnemHead(n)
          else if n>127 then begin
           n:=n-128;
           Str(n,s1);
           s:=s+'_app_'+s1;
          end;
        end;
    'b':begin
         n:=v and MnemBase;
         Str(n,s1);
         s:=s+'_'+s1;
        end; 
    '2':begin
         n:=(v and MnemBase) shr 8;
         Str(n,s1);
         s:=s+'_'+s1;
         n:=(v and MnemBase) and $ff;
         Str(n,s1);
         s:=s+'_'+s1;
        end;
   end;      
  end else Str(v,s);
  GetLong:=s;
 end;
 
procedure NewAtom(s:string; reg:byte);
 var n,i:longint; w1,w2,w3:string;
 procedure CreateInt;
  var i,j,x,y:longint;
  begin
   if w1='none' then Error('undefined SECD code');
   j:=0;
   for i:=1 to MnemMax do if w1=MnemHead(i) then j:=i; 
   if j=0 then Error('Bad mnemonick');
   n:=j shl 24;
   if w2<>'' then begin
    val(w2,x,i);
    if i<>0 then begin
      x:=0;
      for i:=1 to MnemMax do if w2=MnemHead(i) then x:=i; 
      if x=0 then Error('Bad first parameter for SECD code');
      if MnemHead(x)='app' then begin
       val(w3,y,i); w3:='';
       x:=128+y;
      end;
    end;  
    if w3<>'' then begin
     val(w3,y,i);
     if i<>0 then Error('Bad second parameter for SECD code');
     n:=n+256*x+y;
    end else n:=n+x;
   end;
   SetAtom(reg,int,n);
  end;
 begin
  if (s='NIL') or (s='nil') 
   then SetAtom(reg,null,0)
   else begin
    val(s,n,i);
    if i=0
     then SetAtom(reg,int,n)
     else if s[1]='#' then begin
      delete(s,1,1);
      w1:=EatStr(s,'_');
      w2:=EatStr(s,'_');
      w3:=EatStr(s,'_');
      CreateInt;
     end else begin
      n:=AtomPtr-1;
      while (n>=0) and (Atoms[n]<>s) do dec(n);
      if n<0 then begin
       n:=AtomPtr; inc(AtomPtr);
       if AtomPtr>MaxStr then Error('Atom space overflow');
       Atoms[n]:=s;
      end;
      SetAtom(reg,atom,n);
     end;
   end;
 end;
 
procedure ReadSExp(r0,r1:byte; p:PStream);
 var s:string;
 function GetLex:string;
  var ok:boolean;
  begin 
   repeat 
    GetLex:=FGetLex(p,ok);
    if not ok then Delay;
   until ok 
  end;
 procedure ReadTail; forward; 
 procedure ReadSExp0;
  begin
   s:=GetLex;
   if s='(' then ReadTail else NewAtom(s,r0)
  end;
 procedure ReadTail; 
  procedure Tail0;
   begin
    Push(r0,r1); ReadTail; Roll(r1,r0);
   end;
  begin
   s:=GetLex;
   if s='('
    then begin ReadTail; Tail0 end
    else if s=')'
     then SetAtom(r0,null,0) 
     else if s='.' 
      then begin 
       ReadSExp0; s:=GetLex;
       if s<>')' then Error('Bad S-Expresion');
      end 
      else begin NewAtom(s,r0); Tail0 end;
  end;
 begin
  ReadSExp0;
 end;

procedure ReadSFrom(fname:string; r:byte);
 var p:PStream;
 begin
  p:=FOpen(fname,instream);
  ReadSExp(r,MaxReg,p); 
  FClose(p);
 end;
 
procedure FLink(fn:string);
 begin close(stderr); assign(stderr,fn); rewrite(stderr) end;

procedure WriteSTo(fname:string; r:byte);
 begin
  FLink(fname);
  ShowReg(r,99999,true); writeln(stderr);
  FLink('');
 end;
 
begin
 FLink('secd.log');
 AtomPtr:=0;
end.


