unit secd_cod;
interface

const
   s=0;   e=1;   c=2;       { SEC itself     } 
  w1=3;   a=4;   b=5;       { work and cache }
  r0=6;  r1=7;  io=8;       { running lists, io queue }
  sk=9; ek=10; ck=11;       { kernel state }
 s1=12; e1=13; c1=14; u=15; { SEC swap / err_cod } 

 Debug:boolean=false;
 ShdMaxCnt=127;

var
 LoopCnt, SwapCnt,
 CPCnt, 
 PendingCnt, PendMax:longint;
 SwapOK:boolean;
 userMode:boolean;

procedure RTError(n:word; msg:string);
procedure secd_loop;
procedure ShowSECD;

{ Cache memory vars & functions: }

const ChCnt:byte=0;
procedure Just(n:byte);

{ 
  ------------ SECD commands summary: -------------------

  Subcommands:
  
  CleanX:
    repeat
      ([nil.x] s) e c --> (x s) e c
    until S is ([nil.x] s);
    if S is ([p.q] s) then begin
      q --> ((S e c) q)    (dalay current process until recipe [p.q] is not ready) 
      Load pending process from r1;
    end;

 
  Commands:

  RTN (#rtn) --> executed if C=nil
  
    if S is (x [p.q] s) then begin
      CleanX;
      if S is ([c.e] [p.q] s)
        then ([c.e] s) e' nil   --> ([0.nil] s) e c
        else begin
          if x is a node
             then replace [p.q] with x
             else replace [p.q] with [nil.x];
          if x is a node   
             then (x y s) e nil --> (y s) e nil
             else (x y s) e nil --> (x s) e nil;
          while q<>nil do roll(q,r1);
          if p>0 then Primitive(p);   
        end;
    end else (x e' c' s) e nil  --> (x s) e' c'

  1 LD x y       s  e c           --> ((GetNth(GetNth(e,x),y) s) e c  
  2 LDC          s  e (x c)       --> (x s)     e    c 
  3 LDF          s  e (f c)       --> (f.e s)   e    c 
  4 DELAY        s  e (c c')      --> ([c.e] s) e    c' 
  5 CASE_k       if i in [0..k-1] then c_i:=ci else c_i:=ck
                 (i s) e (c0 ... ck c) --> (c_i s)  e c
  6 DUM          s  e c           --> s      (nil e) c 
  7 STOP         Kill the process. Attempt to load some pending process. 
  8 CALL        
         if c=nil
          then   (c' s) e nil     --> s       e c'
          else   (c' s) e c       --> (e c s) e c'

  9 RECV  (Pipe s)   e c     --> Receive message from Pipe 
          (Stream s) e c     --> Receive lexeme from Stream 
          (Connwait s) e c   --> Receive tcp connection from listener or connection try 
 10 SEND  (Pipe msg s) e c   --> Send message to Pipe
          (Stream lex s) e c --> Send lexeme to Stream
 11 RECVC (Stream s) e c     --> Receive char (ascii code) from Stream 
 12 SENDC (Stream I s) e c   --> Send char with ascii code I to Stream
 13 FORK  s e (c1 c2 c)      --> (c1 s) e c
                                 r0 --> (((c2 s) e c) r0)
 14 POP   (x s) e c          --> s e c
 15 PRIM_p <==> Primitive(p) 

 ---------------------------- Binary primitives -----------------------------
  
 16 CONS    (x y s) e c   --> (cons(x,y) s) e c  
 17 EQ      (x y s) e c   --> (eq(x,y)   s) e c
 18
 19
 20
 21
 22 ADD     (x y s) e c   --> (y+x s)       e c 
 23 SUB     (x y s) e c   --> (y-x s)       e c 
 24 MUL     (x y s) e c   --> (y*x s)       e c 
 25 DIV     (x y s) e c   --> ((y div x) s) e c 
 26 REM     (x y s) e c   --> ((y mod x) s) e c 
 27 LEQ     (x y s) e c   --> ((y <= x) s)  e c  --> for integers 
 28 LEXLEQ  (x y s) e c   --> ((y <= x) s)  e c  --> for atom names
 29 
 30 CONCAT  (a b s) e c   --> (a||b s)      e c  --> concatenates atom names
 31
 
 ---------------------------- APP / TR_APP / kernel codes: ---------------------------

 32 LET n        s         e (c' c)    --> (e c Drop(s,n))       (s e)       c' 
 33 TRLET n      s         e c         --> Drop(s,n)             (s e)       c  
 34 LETREC n     s         e (c' c)    --> ((cdr e) c Drop(s,n)) Setcar(e,s) c' 
 35 TRLETR n     s         e c         --> Drop(s,n)             Setcar(e,s) c  
 36 APP n        (c'.e' s) e c         --> (e c Drop(s,n))       (s e')      c' 
 37 TR_APP n     (c'.e' s) e c         --> Drop(s,n)             (s e')      c' 
 38  
 39  
 40 RECOVER (#rec_p - p is unary primitive. Now only #rec_0 is used ! )
    CleanX;
    if S is ([c.e] s)
       then ([c.e] s) e' c' --> ([p.nil] e' c' s) e c
       else (x s) e c       --> (p(x) s) e c   
 
 41 TRREC   (#trrec_p - p is primitive or 
             128+n if encode (... #rec #trapp_n) sequence )
    CleanX;
    if S is ([c.e] s)
       then ([c.e] s) e' c' --> ([p.nil] s) e c
       else (x s) e c       --> (p(x) s) e c  

 42 
 43
 44 USER0
 45 USER1 
 46 INTR n
 47

 ---------------------------- Unary primitives ------------------------------

 48 CAR     (x s) e c  --> (car(x) s)   e c
 49 CDR     (x s) e c  --> (cdr(x) s)   e c
 50 INT     (x s) e c  --> (int?(x) s)  e c
 51 ATOM    (x s) e c  --> (atom?(x) s) e c
 52 NULL    (x s) e c  --> ((x=nil?) s) e c   
 53 NODE    (x s) e c  --> (node?(x) s) e c
 54 
 55 STAT    (x s) e c  --> (x s)        e c   for debuging
 56 INC     (x s) e c  --> (x+1 s)      e c 
 57 DEC     (x s) e c  --> (x-1 s)      e c    
 58 CHR     (n s) e c  --> (chr(n) s)   e c   chr(n) is a one char atom
 59   
 60   
 61
 62
 63
}

implementation

uses secd_mem, secd_str, secd_msg, secd_tcp;

procedure ShowSECD;
 const CMax=80;
 procedure ShowCache;
  begin
   case ChCnt of
    0:begin
       write(stderr,'S=  '); 
      end;
    1:begin
       write(stderr,'Sa= '); ShowReg(a,CMax,false); writeln(stderr);
       write(stderr,'St= '); 
      end;
    2:begin
       write(stderr,'Sa= '); ShowReg(a,CMax,false); writeln(stderr);
       write(stderr,'Sb= '); ShowReg(b,CMax,false); writeln(stderr);
       write(stderr,'St= '); 
      end;
   end; 
   ShowReg(s,CMax,false); writeln(stderr);
  end;
 begin
  writeln(stderr,'SECD: Loops=',LoopCnt,'  Cons=',ConsCnt,
               '   Runs=',PendingCnt); 
  ShowCache;
  write(stderr,'E=  '); ShowReg(e,CMax,false); writeln(stderr);
  write(stderr,'C=  '); ShowReg(c,CMax,false); writeln(stderr);
  {readln;} 
 end;

procedure SwapRegs;
 begin
  SwapReg(s,s1);
  SwapReg(e,e1);
  SwapReg(c,c1);
 end;

procedure Just(n:byte);
 begin if ChCnt<>n then begin
  case ChCnt of
   0:if n=1 
      then Pop(s,a)
      else begin Pop(s,a); Pop(s,b) end;
   1:if n=0 
      then Push(a,s) else Pop(s,b); 
   2:if n=1 
      then Push(b,s)
      else begin Push(b,s); Push(a,s) end;
  end;
  ChCnt:=n;
 end end;

var
 cmd:word;    { current SECK command }
 par:longint; { current command parameter }

procedure TRApp(par:word); 
 begin { TRAPP:   (c'.e' s) e c --> Drop(s,n) (s e')    c' }
  Just(1);
  MoveReg(a,e); 
  Pop(e,c);
  Push(s,e); 
  Drop(s,par);
  ChCnt:=0;
 end; 

procedure Prev(par:word);
 begin
  if ChCnt=2 then Just(1);
  inc(ChCnt);
  if ChCnt=2 then MoveReg(a,b); 
  CopyReg(par+s1,a);
 end;

procedure Intr(par:word);
 begin
  Just(0);
  SetAtom(a,int,par);
  if userMode then begin
   if par<128 
    then CopyReg(a,u)
    else SetAtom(u,int,cmd);
   CopyReg(s1,sk);
   CopyReg(e1,ek);
   CopyReg(c1,ck);
   SwapRegs;
  end else begin
   CopyReg(sk,s);
   CopyReg(ek,e);
   CopyReg(ck,c);
  end;
  ChCnt:=1;
  userMode:=false;
 end;
 
procedure RTError(n:word; msg:string);
 begin
  writeln(stderr);
  writeln(stderr,msg,' !');
  ShowSECD;
  Flush(stderr);
  
  Intr(n+128);
  secd_loop;
 end;

procedure SetBool(b:boolean);
 begin
  if b then SetAtom(a,int,0) else SetAtom(a,int,1);
 end;
 
procedure Kernel1;
 var 
  ok:boolean;
  ws:string;
 begin
  if ChCnt=0 then Just(1);
  if (cmd in [80..86,88]) and (Regs[a].t<>int)
   then RTError(0,'The argument must be int');
  case cmd of
   80:begin
       ws:=FGetLex(PStream(Regs[a].v),ok);
       if ok then NewAtom(ws,w1);
       SetBool(ok); PushR(w1,a);
      end;
   81:begin
       SetAtom(w1,int,ord(FGetChar(PStream(Regs[a].v),ok)));
       SetBool(ok); PushR(w1,a);
      end;
   82:begin
       SetBool(StreamCont(PStream(Regs[a].v)));
       PushR(w1,a);
      end;
   83:SetBool(FEof(PStream(Regs[a].v)));
   84:begin
       FClose(PStream(Regs[a].v));
       SetAtom(a,int,longint(nil));
      end;
   85:Delay(Regs[a].v); 
   86:Intr( Regs[a].v);
   87:SetAtom(a,int,ord(SubTag(a)));
   88:begin
       SetAtom(w1,int,longint(TCPGetConn(PTCPSocket(Regs[a].v),ok)));
       SetBool(ok); PushR(w1,a);
      end;
  end; 
 end;

procedure Kernel2;
 var ws:string;
 begin
  Just(2);
  if (cmd in [65..67,70..72]) and (Regs[a].t<>int)
   then RTError(1,'The second argument must be int');
  if (cmd in [70..72]) and (Regs[b].t<>int)
   then RTError(2,'First argument must be int'); 
  case cmd of
   64:begin
       if Regs[b].t<>int then 
        RTError(5,'First argument must be TStreamType: 0-input 1-output'); 
       case Regs[a].t of
        null:ws:='';
        atom:ws:=GetStr(Regs[a].v);
       end;
       SetAtom(a,int,longint(FOpen(ws,TStreamType(Regs[b].v))));
      end; 
   65:begin
       case Regs[b].t of
        null:ws:='nil';
        int:ws:=GetLong(Regs[b].v);
        atom:ws:=GetStr(Regs[b].v);
        node:RTError(9,'Sending node to stream is impossible');
       end;
       SetBool(FPutStr(PStream(Regs[a].v),ws));
      end;
   66:begin
       if Regs[b].t<>int
        then RTError(10,'SendC allow only integers');
       SetBool(FPutChar(PStream(Regs[a].v),Chr(Regs[b].v)));
      end;
   67:begin
       SetSubTag(b,TSubTag(Regs[a].v));
       MoveReg(b,a);
      end; 
   68:begin SetCar(b,a); MoveReg(b,a) end;
   69:begin SetCdr(b,a); MoveReg(b,a) end;
   70:SetAtom(a,int,longint(
       SOpen(PTCPSocket(Regs[a].v),TStreamType(Regs[b].v))));
   71:SetAtom(a,int,longint(TCPConnect(Regs[a].v,Regs[b].v)));
   72:SetAtom(a,int,longint(TCPListen( Regs[a].v,Regs[b].v)));
  end;     
  ChCnt:=1;
 end;

procedure secd_loop;
 procedure Primitive(cmd:byte);
  procedure UniCmd;
   begin
    if ChCnt=0 then Just(1);
    if (cmd in [56..58]) and (Regs[a].t<>int)
     then RTError(0,'The argument must be int');
    case cmd of
      48:CarSelf(a);
      49:Drop(a,1);
      50:SetBool(Regs[a].t=int);
      51:SetBool(Regs[a].t=atom);
      52:SetBool(Regs[a].t=null);
      53:SetBool(Regs[a].t=node);
      54:;
      55:ShowStat;   
      56:inc(Regs[a].v);
      57:dec(Regs[a].v);
      58:NewAtom(chr(Regs[a].v),a);
      59..63:;
    end;     
   end;
  procedure BinCmd;
   begin
    Just(2);
    if cmd in [22..27] then begin
     if Regs[a].t<>int then RTError(1,'Second argument must be int'); 
     if Regs[b].t<>int then RTError(2,'First argument must be int'); 
    end;
    if cmd in [28,30] then begin
     if Regs[a].t<>atom then RTError(3,'Second argument must be atom'); 
     if Regs[b].t<>atom then RTError(4,'First argument must be atom'); 
    end;
    case cmd of
     16:PushR(b,a);
     17:SetBool((Regs[b].t=Regs[a].t) and (Regs[b].v=Regs[a].v)); 
     18..21:;
     22:Regs[a].v:=Regs[b].v+Regs[a].v;
     23:Regs[a].v:=Regs[b].v-Regs[a].v;
     24:Regs[a].v:=Regs[b].v*Regs[a].v;
     25:Regs[a].v:=Regs[b].v div Regs[a].v;
     26:Regs[a].v:=Regs[b].v mod Regs[a].v;
     27:SetBool(Regs[b].v <= Regs[a].v);
     28:SetBool(GetStr(Regs[b].v) <= GetStr(Regs[a].v));
     30:NewAtom(GetStr(Regs[b].v)+GetStr(Regs[a].v),a);
    end;     
    ChCnt:=1;
   end;
  begin
   if cmd<64 
    then case cmd of
      16..31:BinCmd;
      48..63:UniCmd;
    end
    else if cmd<128
     then Intr(cmd)
     else TRApp(cmd-128);   
  end;
 procedure CleanX;
  var finish:boolean; 
  begin
   repeat
    if TestRec(a) then begin
      Car(a,w1);
      if Regs[w1].t=null 
       then begin Drop(a,1); finish:=false end
       else begin
        finish:=true;
        if Regs[w1].t=int then begin
         Cdr(a,w1);
         Push(e,c);
         Just(0);
         Push(s,c);
         Car(s,a);
         Push(c,w1); 
         SetCdr(a,w1); { [p.q] --> [p.((s e c) q)] } 
         SwapOK:=true;
        end;
       end; 
    end else finish:=true;
   until finish;
  end;
 procedure Rec;
  begin
   if TestRec(a) then begin
    Just(1);
    MoveReg(e,b); 
    Push(c,s);
    StartRec(par);
    ChCnt:=2;
   end else if par>0 then Primitive(par);
  end; 
 procedure TR_Rec;
  begin
   if TestRec(a) then StartRec(par)
   else if par>0 then Primitive(par);
  end;
 procedure Rtn;
  var p:byte; 
  begin
   Just(2);
   if TestRec(b) then begin
    CleanX;
    if SwapOK then Exit;
    if TestRec(a) 
     then StartRec(0)
     else begin
      p:=ReplaceRec;
      { Specialy for nodes the replaced recipe is returned: }
      if Regs[a].t=node then MoveReg(b,a); ChCnt:=1; 
      { activate all processes, waiting for this value }
      while Regs[w1].t=node do begin
       roll(w1,r1); IncPend;
      end; 
      { end of activation }
      if p>0 then begin 
        if p>127 then RTError(6,'#trapp_n into #rtn'); 
        Primitive(p);  
      end;  
     end;
   end else begin
    MoveReg(b,e);
    Pop(s,c);
    ChCnt:=1;
   end;
  end;
 procedure LDx;
  begin
   if ChCnt=2 then Just(1);
   inc(ChCnt);
   if ChCnt=2 then MoveReg(a,b); 
   case cmd of
    1:begin
       GetNth(e,w1,par shr 8);
       GetNth(w1,a,par and $ff);
      end;
    2:Pop(c,a);
    3:begin Pop(c,a); PushR(e,a) end;
    4:begin Pop(c,a); PushR(e,a); SetSubTag(a,recipe) end; 
   end;
  end; 
 procedure CASE_k;
  var k:longint;
  begin
   if ChCnt=0 then Just(1);
   if Regs[a].t=int then k:=Regs[a].v else k:=par;
   if (k<0) or (k>par) then k:=par;
   GetNth(c,a,k);
   Drop(c,par+1);
  end;
 procedure CALL; 
  begin
   if Regs[c].t=node then begin { normal CALL }
    Just(1);
    MoveReg(c,b); 
    MoveReg(a,c);
    CopyReg(e,a);
    ChCnt:=2;
   end else begin               { TR CALL }   
    if ChCnt=0 then Just(1);
    MoveReg(a,c); dec(ChCnt);
    if ChCnt=1 then MoveReg(b,a); 
   end;
  end;
 procedure FORK;
  begin
   Just(0);
   Pop(c,a); 
   CopyReg(s,w1);
   Roll(c,s); 
   CopyReg(c,b);
   ProcSave;
   MoveReg(b,c);
   MoveReg(w1,s);
   Inc(ChCnt);
  end;
 procedure SpecCmd;
  begin
   case cmd of
     1..4:LDx;
     5:CASE_k; 
     6:begin SetAtom(w1,null,-1); Push(w1,e) end;
     7:; { stop - Phase0 }
     8:CALL;
     9:RecvMsg(0);
    10:SendMsg(0);
    11:RecvMsg(1);
    12:SendMsg(1);
    13:FORK;
    14:begin { POP }
        if ChCnt=0 then Just(1);
        dec(ChCnt);
        if ChCnt=1 then MoveReg(b,a); 
       end;
    15:Primitive(par);   
   end;
  end;
 procedure user;
  begin
   ChCnt:=0;
   MoveReg(a,c1);
   MoveReg(b,e1);
   MoveReg(w1,s1);
   SwapRegs;
   SetAtom(u,null,0);
   userMode:=true;
  end;
 procedure CallCmd;
  begin
   case cmd of
    32:begin { LET:     s e (c' c)  --> (e c Drop(s,n)) (s e) c'  }
        Just(0);
        Cdr(c,b);
        CopyReg(e,a); 
        Push(s,e);
        Drop(s,par);
        CarSelf(c);
        ChCnt:=2;
       end;
    33:begin { TRLET:   s e c       --> Drop(s,n) (s e) c       }
        Just(0); 
        Push(s,e); 
        Drop(s,par);
       end;
    34:begin { LETREC:  s e (c' c)  --> ((cdr e) c Drop(s,n)) Setcar(e,s) c' }
        Just(0); 
        Cdr(c,b);
        Cdr(e,a); 
        SetCar(e,s);
        Drop(s,par);
        CarSelf(c);
        ChCnt:=2;
       end;
    35:begin { TRLETR:  s e c       --> Drop(s,n) Setcar(e,s) c  } 
        Just(0); 
        SetCar(e,s); 
        Drop(s,par);
       end;
    36:begin { APP:     (c'.e' s) e c --> (e c Drop(s,n)) (s e') c' }
        Just(1);
        MoveReg(c,b); 
        MoveReg(a,c); 
        MoveReg(e,a); 
        Cdr(c,e); 
        Push(s,e);
        CarSelf(c);
        Drop(s,par);
        ChCnt:=2;
       end; 
    37:TRApp(par); { TRAPP:   (c'.e' s) e c --> Drop(s,n) (s e') c' }
    38:;
    39:;
    40:Rec;
    41:TR_Rec;
    42:;
    43:;
    44:begin Just(2); Pop(s,w1); user end; 
    45:begin Just(1); Pop(a,w1); Pop(a,b); CPCnt:=0; user end;
    46:Intr(par);
    47:Prev(par);
   end;
  end;
 procedure Phase0;
  begin
   case cmd of
     7:SwapOK:=true;
    40,41:begin { Rec / TR_Rec }
       if ChCnt=0 then Just(1);
       CleanX;
      end;
   end;
  end;
 procedure Phase1;
  begin
   Drop(c,1); 
   case cmd of
     1..15:SpecCmd;
    16..31:Primitive(cmd);
    32..47:CallCmd;
    48..63:Primitive(cmd);
    64..79:Kernel2;
    80..95:Kernel1;
   end;
  end;
 begin
  SwapOK:=false;
  repeat
   if Debug then ShowSECD; 
   inc(LoopCnt);
   if userMode then inc(CPCnt);

   { Test if timelimit is reached and if there is another 
     running or sleaping processes: }  
   if (CPCnt>ShdMaxCnt) and 
      userMode and
      ((Regs[r0].t=node) or
       (Regs[r1].t=node) or
       (Regs[io].t=node)) 
    then begin
     inc(SwapCnt);
     Just(0);
     ProcSave;
     IO_Scan;
     SwapOK:=true;
     if debug then writeln(stderr,'----------- SWAP -------------');
    end
    else if Regs[c].t=null 
     then Rtn
     else begin
      Car(c,w1);
      if Regs[w1].t<>int 
       then RTError(7,'Command not found at the top of C') 
       else begin
        cmd:=Regs[w1].v shr 24;
        par:=Regs[w1].v and $ffffff;
        Phase0;
        if not SwapOK then Phase1;
       end;
     end;

   if SwapOK then ProcLoad;
   
  until SwapOK;
  Just(0);
 end;

begin
 SwapCnt:=0;
 LoopCnt:=0;
 PendingCnt:=1;
 PendMax:=PendingCnt;
 CPCnt:=0;
end.


