unit secd_mem;
interface

const
 MaxReg=19;
type
 TTag = (null,node,int,atom);
 TSubTag = (cons, recipe, pipe, stream, connwait);
 TReg = record
  t:TTag;
  v:longint
 end;
 TNode = record
  s:TSubTag;
  m:boolean; { Mark/Release flag }
  c:longint; { Reference counter }
  l,r:TReg;
 end;

var
 Regs: array[0..MaxReg] of TReg;
 ConsCnt:longint;
 NodePtr:longint;

procedure SetAtom(reg:byte; t:TTag; v:longint);
procedure ShowReg(reg:byte; limit:longint; lnCase:boolean);
procedure ShowStat;
procedure Error(s:string);

function  SubTag(s:byte):TSubTag;
procedure SetSubTag(s:byte;t:TSubTag);
function  TestRec(s:byte):boolean;
procedure CreateIO(i:longint;t:TSubTag);

procedure StartRec(p:byte);
procedure Car(s,d:byte);
procedure CarSelf(s:byte);
procedure Cdr(s,d:byte);
procedure GetNth(s,d:byte; i:word);
procedure Drop(s:byte; i:word);
procedure Pop(s,d:byte);
{
procedure Cons(l,r,d:byte);
procedure CdrSelf(s:byte);
}
procedure Push (s,d:byte);
procedure PushR(s,d:byte);
procedure Roll(s,d:byte);
procedure CopyReg(s,d:byte);
procedure MoveReg(s,d:byte);
procedure SwapReg(s,d:byte);
procedure SetCar(s,d:byte);
procedure SetCdr(s,d:byte);
function  ReplaceRec:byte;

implementation

uses secd_str, secd_cod;

const
 MaxNode = 200000;
var
 Nodes:array[0..MaxNode] of TNode;

{---------- Common Garbage Collector instruments ---------}

const
 GCMax=MaxNode div 3; { Garbage Collector Stack size }
var 
 GCStack:array[0..GCMax] of longint;
 GCStPtr:word;
 GCSPMax:word;

procedure AddNode(v:longint; var Avail:TReg; var AvCnt:longint);
 begin
  Nodes[v].l:=Avail;
  Nodes[v].s:=cons;
  Avail.t:=node;
  Avail.v:=v;
  inc(AvCnt);
 end;

procedure GCPush(v:longint);
 begin
  GCStack[GCStPtr]:=v;
  if GCStPtr<=GCMax 
   then inc(GCStPtr)
   else Error('Garbage collector stack overflow');
  if GCSPMax<GCStPtr then GCSPMax:=GCStPtr; 
 end;

{---------- Reference Count Garbage Collector ------------}

var 
 RCAvail:TReg;
 RCAvCnt:longint;
 RCMax:longint;

procedure RC_Collect(v1:longint);
 var w:TReg; v:longint; finish:boolean;
 begin
  v:=v1;
  repeat
  
   { Delay release of right branch of the tree }
   w:=Nodes[v].r;
   if w.t=node then GCPush(w.v);
   
   { Immediate release of left branch of the tree }
   w:=Nodes[v].l;

   AddNode(v,RCAvail,RCAvCnt); 
   
   finish:=true;
   if w.t=node then begin
    v:=w.v;
    dec(Nodes[v].c);
    finish:=Nodes[v].c>0;
   end;
   
  until finish; 
 end;

procedure RC_inc(w:TReg);
 var i:longint;
 begin
  if w.t=node then begin
   inc(Nodes[w.v].c);
   i:=Nodes[w.v].c;
   if i>RCMax then RCMax:=i;
  end; 
 end;

procedure RC_dec0(v:longint);
 begin
  dec(Nodes[v].c);
  if Nodes[v].c=0 then RC_Collect(v);
  if Nodes[v].c<0 then Error('RefCount<0');
 end;

procedure RC_dec(w:TReg);
 begin
  if w.t=node then RC_dec0(w.v)
 end;

{------------ Mark/Release Garbage Collector -------------}

var
 MRAvail:TReg;
 MRAvCnt:longint;

procedure MarkReg(w:TReg);
 var v:longint;
 procedure Mark0(w:TReg);
  procedure Mark1(v:longint);
   begin if not Nodes[v].m then begin 
     Nodes[v].m:=true; GCPush(v) 
   end end;
  begin if w.t=node then Mark1(w.v) end; 
 begin 
  Mark0(w);
  while GCStPtr>0 do begin
   dec(GCStPtr);
   v:=GCStack[GCStPtr];
   Mark0(Nodes[v].l);
   Mark0(Nodes[v].r);
  end;
 end;

procedure Release_GC;
 var v:longint;
 begin
  write(stderr,' OK, ');
  for v:=0 to MaxNode do if Nodes[v].m
   then Nodes[v].m:=false
   else AddNode(v,MRAvail,MRAvCnt); 
  writeln(stderr,MRAvCnt,' nodes released. '); 
  Flush(stderr);
 end;

procedure MR_GC;
 var i:byte;
 begin
  write(stderr,'Mark '); 
  for i:=0 to MaxReg do begin
   MarkReg(Regs[i]);
   write(stderr,'.');
  end; 
  Release_GC;
 end;
 
{--------- GetNewNode / Garbage Collect initiator --------}

procedure GetNewNode(var v:longint);
 begin
  { Force new RefCount Collector steps if RCAvail=nil }
  while (RCAvail.t<>node) and (GCStPtr>0) do begin
   dec(GCStPtr);
   RC_dec0(GCStack[GCStPtr]);
  end;
  
  if RCAvail.t=node then begin { RC node avialable }
   dec(RCAvCnt);
   v:=RCAvail.v;
   RCAvail:=Nodes[v].l;
  end else                        { no RC node avialable; GCStPtr=0 }
   if NodePtr>MaxNode then begin  { no free Heap, Mark/Release usage } 
    if MRAvail.t=null then MR_GC; { Mark/Release Garbage Collect }
    if MRAvail.t=node then begin  { MR node avialable }
     dec(MRAvCnt);
     v:=MRAvail.v;
     MRAvail:=Nodes[v].l;
    end else Error('Node space overflow');
   end else begin                 { get node from the Heap space } 
    v:=NodePtr; inc(NodePtr);
    Nodes[v].s:=cons;
    Nodes[v].m:=false;
   end; 
  inc(ConsCnt);
 end;

{---------- List-oriented meta_commands ------------------}

function SubTag(s:byte):TSubTag;
 begin
  if Regs[s].t=node
   then SubTag:=Nodes[Regs[s].v].s
   else RTError(19,'SubTag of an atom');
 end;
procedure SetSubTag(s:byte;t:TSubTag);
 begin
  if Regs[s].t=node then begin
   Nodes[Regs[s].v].s:=t;
  end else RTError(20,'SetSubTag to an atom');
 end; 
function TestRec(s:byte):boolean;
 begin
  if Regs[s].t=node
   then TestRec:=SubTag(s)=recipe
   else TestRec:=false;
 end;
 
procedure CreateIO(i:longint;t:TSubTag);
 begin
  SetAtom(a,int,i);
  SetAtom(b,null,0);
  PushR(b,a);
  SetSubTag(a,t);
 end;
                             
procedure Cdr(s,d:byte);
 var w:TReg;
 begin 
  if s<>d then begin
   w:=Regs[s];
   if w.t=node then begin
    RC_dec(Regs[d]);
    Regs[d]:=Nodes[w.v].r;
    RC_inc(Regs[d]);
   end else RTError(21,'Cdr of an atom')
  end else Error('Use Drop(x,1)')
 end;

procedure CarSelf(s:byte);
 var w:TReg;
 begin 
  w:=Regs[s];
  if w.t=node then begin
   Regs[s]:=Nodes[w.v].l;
   RC_inc(Regs[s]);
   RC_dec(w);
  end else RTError(22,'CarSelf of an atom')
 end;
 
procedure Drop(s:byte; i:word);
 var j:word; w:TReg;
 begin if i>0 then begin
  j:=i;
  w:=Regs[s];
  while j>0 do begin
   if w.t=node 
    then w:=Nodes[w.v].r
    else RTError(23,'List for Drop too short / CdrSelf of an atom');
   dec(j); 
  end;
  RC_inc(w);
  RC_dec(Regs[s]);
  Regs[s]:=w;
 end end;

{ procedure CdrSelf(s:byte); begin Drop(s,1) end; }

procedure GetNth(s,d:byte; i:word);
 var j:word; w:TReg;
 begin
  if s<>d then begin
   j:=i;
   w:=Regs[s];
   while j>0 do begin
    if w.t=node 
     then w:=Nodes[w.v].r
     else RTError(24,'List for GetNth too short');
    dec(j); 
   end;
   RC_dec(Regs[d]);
   if w.t=node
    then Regs[d]:=Nodes[w.v].l
    else RTError(25,'i-th element is atom in GetNth / car of an atom');
   RC_inc(Regs[d]);
  end else Error('GetNth(s,s,i) is impossible');
 end;

procedure Car(s,d:byte); begin GetNth(s,d,0) end;

procedure Pop(s,d:byte); begin GetNth(s,d,0); Drop(s,1) end; 

procedure Push_0(d:byte; l,r:TReg; normal:boolean);
 var v:longint; 
 begin
  GetNewNode(v);
  Nodes[v].l:=l;
  Nodes[v].r:=r;
  if normal
   then RC_inc(l)
   else RC_inc(r);
  Nodes[v].c:=1;
  Regs[d].t:=node;
  Regs[d].v:=v;
 end;
  
procedure Push (s,d:byte); 
 begin
  if s<>d 
   then begin Push_0(d,Regs[s],Regs[d],true) end
   else Error('Push(x,x) is impossible'); 
 end;

procedure PushR(s,d:byte); 
 begin
  if s<>d
   then begin Push_0(d,Regs[d],Regs[s],false) end
   else Error('PushR(x,x) is impossible'); 
 end;

procedure Roll(s,d:byte);
 var w:TReg; 
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   w:=Nodes[w.v].l;
   Push_0(d,w,Regs[d],true);
   Drop(s,1);
  end else RTError(26,'Roll of an atom')
 end else Error('Roll(x,x) is impossible') 
 end;
 
procedure CopyReg(s,d:byte);
 begin 
  if s<>d then begin
   RC_dec(Regs[d]);
   Regs[d]:=Regs[s];
   RC_inc(Regs[d]);
  end else Error('CopyReg(s,s) is impossible');  
 end;

procedure MoveReg(s,d:byte);
 begin 
  if s<>d then begin
   RC_dec(Regs[d]);
   Regs[d]:=Regs[s];
   Regs[s].t:=null; 
   Regs[s].v:=0;
  end else Error('MoveReg(s,s) is impossible');  
 end;

procedure SwapReg(s,d:byte);
 var w:TReg;
 begin w:=Regs[s]; Regs[s]:=Regs[d]; Regs[d]:=w end;
 
procedure SetCar(s,d:byte);
 var w,o:TReg;
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   o:=Nodes[w.v].l;
   Nodes[w.v].l:=Regs[d];
   RC_inc(Regs[d]);
   RC_dec(o);
  end else RTError(27,'SetCar to an atom')
 end else Error('SetCar(x,x) is impossible')
 end;

procedure SetCdr(s,d:byte);
 var w,o:TReg;
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   o:=Nodes[w.v].r;
   Nodes[w.v].r:=Regs[d];
   RC_inc(Regs[d]);
   RC_dec(o);
  end else RTError(28,'SetCdr to an atom')
 end else Error('SetCdr(x,x) is impossible')
 end;

function ReplaceRec:byte;
 var w:TReg; v,v2:longint;
 begin 
  if TestRec(b) then begin 
    v:=Regs[b].v;
    w:=Nodes[v].l;
    if w.t=int then begin
     ReplaceRec:=w.v;
     RC_dec(Regs[w1]);
     Regs[w1]:=Nodes[v].r;
     if Regs[a].t=node then begin
      v2:=Regs[a].v;
      w:=Nodes[v2].l;
      RC_inc(w);
      Nodes[v].l:=w;
      w:=Nodes[v2].r;
      RC_inc(w);
      Nodes[v].r:=w;
      Nodes[v].s:=SubTag(a); 
     end else begin
      SetAtom(MaxReg,null,0);
      Nodes[v].l:=Regs[MaxReg];
      Nodes[v].r:=Regs[a];
     end; 
    end else RTError(29,'Replace to non-working recipe'); 
  end else RTError(30,'Replace to non-recipe data');
 end;

procedure StartRec(p:byte);
 var v:longint;
 begin
  if Regs[a].t=node then begin
   v:=Regs[a].v;
   RC_dec(Regs[c]);
   Regs[c]:=Nodes[v].l;
   RC_dec(Regs[e]);
   Regs[e]:=Nodes[v].r;
   SetAtom(MaxReg,null,0);
   Nodes[v].r:=Regs[MaxReg];
   SetAtom(MaxReg,int,p);
   Nodes[v].l:=Regs[MaxReg];
  end else RTError(31,'StartRec to an atom');
 end;  
 
procedure SetAtom(reg:byte; t:TTag; v:longint);
 begin if t<>node then begin
  RC_dec(Regs[reg]);
  Regs[reg].t:=t; Regs[reg].v:=v;
 end end;

{------------------------ Misc ---------------------------}

procedure Error(s:string);
 begin
  writeln(stderr);
  writeln(stderr,s,' !');
  ShowSECD;
  Flush(stderr);
  Close(stderr); 
  Halt;
 end;
procedure ShowStat;
 begin
  write  (stderr,'--- SECD : '); 
  write  (stderr,'ConsUse=',ConsCnt:7,' ');
  write  (stderr,'  SECD_loops=',LoopCnt:7,' ');
  writeln(stderr,'  AtomUse=',AtomPtr:5,' ');
  write  (stderr,'--- stats: '); 
  write  (stderr,'HeapUse=',NodePtr-MRAvCnt:7,' ');
  write  (stderr,'  RefCntMax= ',RCMax:7,' ');
  writeln(stderr,'  GCSPMax=',GCSPMax:5,' ');
  write  (stderr,'---------: '); 
  write  (stderr,'SwapCnt=',SwapCnt:7,' ');
  write  (stderr,'  ShdMaxCnt= ',ShdMaxCnt:7,' ');
  writeln(stderr,'  PendMax=',PendMax:5,' ');
  Flush(stderr);
 end;

{---------------------------------------------------------}

procedure ShowReg(reg:byte; limit:longint; lnCase:boolean);
 var cnt,lcnt:longint; depth,d0:word;
 procedure wr(s:string);
  var i:byte;
  begin
   for i:=1 to length(s) do begin
    inc(cnt); inc(lcnt);
    if cnt<limit then write(stderr,s[i]);
   end;
  end;
 procedure ShowReg_0(R:TReg); forward;
 procedure ShowTail(R:TReg);  forward;
 procedure ShowAtom(R:TReg);
  var s:string;
  begin
   case R.t of
    int:s:=GetLong(R.v);
    atom:s:=GetStr(R.v);
   end;
   wr(s);
  end;
 procedure ShowNode(v:longint);
  var w:TReg;
  begin
   w:=Nodes[v].l;
   ShowReg_0(w);
   w:=Nodes[v].r;
   ShowTail(w);
  end;
 procedure ShowTail(R:TReg);
  begin if cnt<limit then
   case R.t of
    node:begin wr(' '); ShowNode(R.v) end;
    int,atom:begin wr('.'); ShowAtom(R) end;
    null:;
   end; 
  end;
 procedure ShowReg_0(R:TReg);
  var i:word;
  begin if cnt<limit then 
   case R.t of
    null:wr('NIL');
    node:begin 
          if (lcnt-3*(depth-d0)>40) and lnCase then begin
           writeln(stderr); lcnt:=0; d0:=depth;
           for i:=0 to depth do wr(' ');
          end;
          if Nodes[R.v].s=cons then wr('(') else wr('[');
          inc(depth); ShowNode(R.v); dec(depth); 
          if Nodes[R.v].s=cons then wr(')') else wr(']');
         end;
    int,atom:ShowAtom(R);
   end;
  end;
 begin
  cnt:=0; lcnt:=0; depth:=0; d0:=0;
  ShowReg_0(Regs[reg]) 
 end;
 
procedure InitMem;
 var i:byte;
 begin
  for i:=0 to MaxReg do Regs[i].t:=null;
  NodePtr:=0; ConsCnt:=0;
  RCAvCnt:=0; RCAvail.t:=null; 
  GCStPtr:=0; GCSPMax:=0; RCMax:=0;
  MRAvCnt:=0; MRAvail.t:=null;
 end;

begin
 InitMem;
end.

