unit secd_tcp;

interface

uses Sockets, Linux;

type
 PTCPSocket = ^TTCPSocket;
 TCPConnTyp  = (listener, conn_from, conn_to);
 TTCPSocket = record
  conn:longint;
  SA:TInetSockAddr;
  t:TCPConnTyp;
  pcnt:word;
  broken:boolean;
 end;

function TCPListen (aPort:word; maxconn:word):PTCPSocket;
function TCPConnect(aPort:word; host:longint):PTCPSocket;
function TCPGetConn(socket:PTCPSocket; var ok:boolean):PTCPSocket;

implementation

uses secd_mem, secd_cod;

const
 Err_TryAgain=11;
 Err_InProgress=115;
 
function SwapWord(w:word):word;
 begin
  SwapWord:=(w shl 8) or (w shr 8);
 end;
    
function TCPSocket(aPort:word):PTCPSocket;
 var p:PTCPSocket;
 begin
  New(p);
  with p^ do begin
   conn:=Socket(AF_INET,SOCK_STREAM,0 {6} {PF_INET});
   if conn<0 then RTError(14,'socket');
   fcntl(conn,F_SetFl,Open_NonBlock);
   with SA do begin
    family:=AF_INET;
    port:=SwapWord(aPort);
   end; 
  end; 
  TCPSocket:=p;
 end;
 
function TCPListen (aPort:word; maxconn:word):PTCPSocket;
 var p:PTCPSocket;
 begin
  p:=TCPSocket(aPort);
  with p^ do begin
   SA.addr:=0; { allow connections from anywhere }
   t:=listener;
   if not Bind(conn,SA,SizeOf(SA)) then RTError(15,'bind');
   if not Listen(conn,maxconn)     then RTError(16,'listen');
  end;  
  TCPListen:=p;
 end;
 
function TCPConnect(aPort:word; host:longint):PTCPSocket;
 var p:PTCPSocket;
 begin
  p:=TCPSocket(aPort);
  with p^ do begin
   SA.addr:=host; { ((1 shl 24) or 127) for localhost } 
   t:=conn_to;
  end;  
  TCPConnect:=p;
 end;
 
function TCPGetConn(socket:PTCPSocket; var ok:boolean):PTCPSocket;
 var p:PTCPSocket; addrlen:longint;
 begin 
  case socket^.t of
   listener:begin
     New(p);
     p^.t:=conn_from;
     addrlen:=SizeOf(p^.SA);
     p^.conn:=Accept(socket^.conn,p^.SA,addrlen);
     p^.pcnt:=0;
     p^.broken:=false;
     ok:=p^.conn>=0;
     {
     begin
      write(stderr,' al=',addrlen);
      write  (stderr,' s=',longint(socket),' p=',longint(p));
      writeln(stderr,
        ' SE=',SocketError,' LE=',LinuxError,' conn=',p^.conn,
        ' socket=',socket^.conn);
      flush(stderr);   
     end;
     }
     if ok 
      then begin
       TCPGetConn:=p;
       writeln(stderr,'TCP fd=',p^.conn); flush(stderr);
      end 
      else if SocketError<>Err_TryAgain 
       then RTError(17,'accept')
       else Dispose(p);     
    end;
   conn_to:begin
     with socket^ do begin
      ok:=Connect(conn,SA,SizeOf(SA));
      pcnt:=0;
      broken:=false;
     end;
     if (not ok) and (SocketError<>Err_InProgress) then begin
      writeln(stderr,'SocketError=',SocketError);
      RTError(18,'connect');
     end;
     TCPGetConn:=socket;
    end;
  end;  
 end;

end.

