unit secd_str;
interface

uses secd_tcp;

var 
 AtomPtr: longint;

function GetStr(v:longint):string;
function GetLong(v:longint):string;
procedure NewAtom(s:string; reg:byte);

procedure ReadSFrom(fname:string; r:byte);
procedure WriteSTo (fname:string; r:byte);

type
 TAtomS=string[32];
 PStream = ^TStream;
 TStreamType = (instream, outstream);
 TFileDescrT = (stdio, regfile, tcpconn);
 TStream = record
  t:TStreamType;
  sp:PTCPSocket;
  fd:longint;
  ef:boolean;
  fdt:TFileDescrT;
  lex:TAtomS;
  w:char;
  srmode:byte;
 end;

function  FOpen(fn:string;         st:TStreamType):PStream;
function  SOpen(socket:PTCPSocket; st:TStreamType):PStream;
procedure FClose(p:PStream);
function  FGetLex (p:PStream; var ok:boolean):string;
function  FGetChar(p:PStream; var ok:boolean):char;
function  FPutStr(p:PStream; s:string):boolean;
function  FPutChar(p:PStream; c:char):boolean;
function  FEof(p:PStream):boolean;
function  StreamCont(ps:PStream):boolean;

procedure Delay(n:word);

implementation

{ ---------- OS dependant code --------- }

uses secd_cod, secd_mem, Linux, Sockets;

const 
 SpecChar=['(',')','.',' '];
 MaxStr = 500;
var
 Atoms: array[0..MaxStr] of TAtomS;

const
 Err_TryAgain = 11;

function NewStream(
          aST:TStreamType;
          aSP:PTCPSocket;
          aFD:longint;
          aFDT:TFileDescrT):PStream;
 var p:PStream;
 begin
  New(p);
  with p^ do begin
   t:=aST;
   ef:=false;
   lex:='';
   w:=' ';
   fdt:=aFDT;
   sp :=aSP;
   fd :=aFD;
   fcntl(fd,F_SetFl,Open_NonBlock);
  end;
  NewStream:=p;
 end;

function  FOpen(fn:string; st:TStreamType):PStream;
 begin case st of
  instream: if fn=''
   then FOpen:=NewStream(st,nil,0,stdio)
   else FOpen:=NewStream(st,nil,fdOpen(fn,Open_RdOnly),regfile);
  outstream: if fn=''
   then FOpen:=NewStream(st,nil,1,stdio)
   else begin
    UnLink(fn);
    FOpen:=NewStream(st,nil,fdOpen(fn,Open_WrOnly or Open_Creat),regfile);
   end; 
 end end;

function  SOpen(socket:PTCPSocket; st:TStreamType):PStream;
 begin 
  SOpen:=NewStream(st,socket,socket^.conn,tcpconn);
  inc(socket^.pcnt);
 end;

procedure CTest(p:PStream);
 begin
  if p=nil then RTError(32,'File is allready closed'); 
 end;
 
procedure FClose(p:PStream);
 begin
  CTest(p);
  with p^ do case fdt of
   stdio:;
   regfile:fdClose(fd);
   tcpconn:begin
     dec(sp^.pcnt);
     if sp^.pcnt=0 then begin
      shutdown(fd,2);
      fdClose(fd);
      Dispose(sp);
     end;
    end;
  end; 
  Dispose(p);
 end;

function  FEof(p:PStream):boolean;
 begin
  CTest(p);
  FEof:=false;
  with p^ do case fdt of
   stdio:;
   regfile:if t=instream then FEof:=ef;
   tcpconn:FEof:=sp^.broken;
  end; 
 end;

procedure SetEof(p:PStream);
 begin with p^ do case fdt of
  regfile:ef:=true;
  tcpconn:sp^.broken:=true;
 end end;

procedure fdRdErr;
 begin
  writeln(stderr,'LinuxError=',LinuxError);
  RTError(33,'fdRead')
 end;
      
function  FGetLex(p:PStream; var ok:boolean):string;
 var Ready:boolean; ow:char; 
 procedure rd;
  begin with p^ do begin
   Ready:=true;
   if FEof(p) 
    then w:=')' 
    else if fdRead(fd,w,1)<>1  
     then case LinuxError of
      0:begin SetEof(p); w:=' ' end;
      Err_TryAgain:Ready:=false;
      else fdRdErr;
     end; 
   if (w=chr(10)) or (w=chr(13)) then w:=' ';
  end end;
 begin 
  CTest(p);
  with p^ do
   if t=outstream then RTError(34,'recv from output stream')
   else begin
    srmode:=0;
    ok:=false; Ready:=true;
    while Ready and (not ok) do 
     if (lex='') and (w=' ') then rd else 
      if w in SpecChar
       then if lex=''
        then begin 
         ow:=w; rd;
         if Ready then begin 
          FGetLex:=ow; 
          ok:=true 
         end; 
        end
        else begin 
         FGetLex:=lex; 
         ok:=true; lex:='' 
        end
       else begin 
        ow:=w; rd;
        if Ready then lex:=lex+ow; 
       end
   end;
 end;

function  FGetChar(p:PStream; var ok:boolean):char;
 begin 
  CTest(p);
  with p^ do
   if t=outstream then RTError(35,'recvc from output stream')
   else begin
    srmode:=1;
    ok:=true;
    if FEof(p) 
     then w:=')' 
     else if fdRead(fd,w,1)<>1  
      then case LinuxError of
       0:begin SetEof(p); w:=' ' end;
       Err_TryAgain:ok:=false;
       else fdRdErr;
      end; 
    FGetChar:=w; 
   end;
 end; 

procedure fdWrErr;
 begin
  writeln(stderr,'LinuxError=',LinuxError);
  RTError(36,'fdWrite')
 end;
      
function FPutStr(p:PStream; s:string):boolean;
 var n,m:longint;
 begin 
  CTest(p);
  with p^ do if t=instream 
   then RTError(37,'send to input stream')
   else begin
    srmode:=0;
    n:=length(s);
    if FEof(p)
     then FPutStr:=true
     else begin
      m:=fdWrite(fd,s[1],n);
      if m=n 
       then FPutStr:=true
       else case LinuxError of
        0:SetEof(p); 
        Err_TryAgain:begin
          FPutStr:=false;
          lex:=copy(s,m+1,n-m);
         end; 
        else fdWrErr;
       end;
      end; 
     end;  
 end;

function  FPutChar(p:PStream; c:char):boolean;
 begin 
  CTest(p);
  with p^ do if t=instream 
   then RTError(39,'sendc to input stream')
   else begin
    srmode:=1;
    if FEof(p)
     then FPutChar:=true
     else if 1=fdWrite(fd,c,1)
      then FPutChar:=true
      else case LinuxError of
       0:SetEof(p);
       Err_TryAgain:begin FPutChar:=false; w:=c end; 
       else fdWrErr;
      end;
   end;  
 end;

function StreamCont(ps:PStream):boolean;
 var 
  ok:boolean; 
  ws:string;
  wc:char;
 begin 
  case ps^.t of 
   instream:case ps^.srmode of
     0:begin
        ws:=FGetLex(ps,ok);
        if ok then NewAtom(ws,w1); { w1 = GetLex }
       end;
     1:begin
        wc:=FGetChar(ps,ok);
        if ok then SetAtom(w1,int,ord(wc)); { w1 = ord(GetChar) }
       end;
    end;
   outstream:begin
     case ps^.srmode of
      0:ok:=FPutStr(ps,ps^.lex);
      1:ok:=FPutChar(ps,ps^.w);
     end;
     if ok then SetAtom(w1,int,0); { w1 = true }
    end; 
  end;
  StreamCont:=ok;
 end;

procedure Delay(n:word);
 begin
  Select(0,nil,nil,nil,n);
 end;

{ ---------- end of OS dependant code --------- }

function GetStr(v:longint):string;
 begin GetStr:=Atoms[v] end;

const
 MnemMax=95;
 SECDmnem:array[1..MnemMax] of TAtomS=
 (
  {
    RTN code is represented by null C register !!!
  }
  { 01..15 - Special codes }
  'ld-2',
  'ldc',
  'ldf',
  'delay',
  'case-b',
  'dum',
  'stop',
  
  'call',
  'none',
  'none',
  'none',
  'none',
  'none',
  'pop',
  'prim-b',
  
  { 16..31 - Binary primitives }
  'cons',
  'eq',
  'none',
  'none',
  'none',
  'none',
  'add',
  'sub',
  
  'mul',
  'div',
  'rem',
  'leq',
  'lexleq',
  'none',
  'concat',
  'none',

  { 32..47 - Calls }  
  'let-b',
  'trlet-b',
  'letrec-b',
  'trletr-b',
  'app-b',
  'trapp-b',
  'none',
  'none',
  
  'rec-p',
  'trrec-p',
  'none',
  'none',
  'user0',
  'user1',
  'intr-b',
  'prev-b',

  { 48..63 - Unary primitives }
  'car',
  'cdr',
  'int',
  'atom',
  'null',
  'node',
  'none',
  'stat',

  'inc',
  'dec',
  'chr',
  'none',
  'none',
  'none',
  'none',
  'none',

  { 64..79 - Kernel2 }  
  'openf',
  'putlex',
  'putc',
  'settag',
  'setcar',
  'setcdr',
  'opens',
  'tcpconn',
  
  'tcplsn',
  'none',
  'none',
  'none',
  'none',
  'none',
  'none',
  'none',

  { 80..95 - Kernel1 }
  'getlex',
  'getc',
  'scont',
  'seof',
  'close',
  'pause',
  'pintr',
  'gettag',
  
  'getconn',
  'none',
  'none',
  'none',
  'none',
  'none',
  'none',
  'none'
 ); 
 MnemBase=$FFFFFF;


function EatStr(var s:string; dl:char):string;
 var n:byte;
 begin
  n:=pos(dl,s);
  if n>0 then begin
   EatStr:=copy(s,1,n-1);
   delete(s,1,n);
   while (length(s)>0) and (s[1]=dl) do delete(s,1,1);
  end else begin
   EatStr:=s;
   s:='';
  end;
 end;

function MnemHead(i:byte):string;
 var c1:string;
 begin
  c1:=SECDmnem[i];
  MnemHead:=EatStr(c1,'-');
 end; 
function MnemTail(i:byte):string;
 var c1:string;
 begin
  c1:=SECDmnem[i];
  EatStr(c1,'-');
  MnemTail:=c1;
 end; 

function GetLong(v:longint):string;
 var s,s1,mt:string; n:longint;
 begin
  if v>MnemBase then begin
   n:=v shr 24;
   if n>MnemMax then RTError(40,'Big SECD code');
   s:='#'+MnemHead(n);
   mt:=MnemTail(n);
   if mt<>'' then case mt[1] of
    'p':begin
         n:=v and MnemBase;
         if n in [1..127] 
          then if n<64
            then s:=s+'_'+MnemHead(n)
            else begin
             Str(n,s1);
             s:=s+'_'+s1;
            end
          else if n>127 then begin
           n:=n-128;
           Str(n,s1);
           s:=s+'_app_'+s1;
          end;
        end;
    'b':begin
         n:=v and MnemBase;
         Str(n,s1);
         s:=s+'_'+s1;
        end; 
    '2':begin
         n:=(v and MnemBase) shr 8;
         Str(n,s1);
         s:=s+'_'+s1;
         n:=(v and MnemBase) and $ff;
         Str(n,s1);
         s:=s+'_'+s1;
        end;
   end;      
  end else Str(v,s);
  GetLong:=s;
 end;
 
procedure NewAtom(s:string; reg:byte);
 var n,i:longint; w1,w2,w3:string;
 procedure CreateInt;
  var i,j,x,y:longint;
  begin
   if w1='none' then RTError(41,'undefined SECD code');
   j:=0;
   for i:=1 to MnemMax do if w1=MnemHead(i) then j:=i; 
   if j=0 then RTError(42,'Bad SECD mnemonick');
   n:=j shl 24;
   if w2<>'' then begin
    val(w2,x,i);
    if i<>0 then begin
      x:=0;
      for i:=1 to MnemMax do if w2=MnemHead(i) then x:=i; 
      if x=0 then RTError(43,'Bad first parameter for SECD code');
      if MnemHead(x)='app' then begin
       val(w3,y,i); w3:='';
       x:=128+y;
      end;
    end;  
    if w3<>'' then begin
     val(w3,y,i);
     if i<>0 then RTError(44,'Bad second parameter for SECD code');
     n:=n+256*x+y;
    end else n:=n+x;
   end;
   SetAtom(reg,int,n);
  end;
 begin
  if (s='NIL') or (s='nil') 
   then SetAtom(reg,null,0)
   else begin
    val(s,n,i);
    if i=0
     then SetAtom(reg,int,n)
     else if s[1]='#' then begin
      delete(s,1,1);
      w1:=EatStr(s,'_');
      w2:=EatStr(s,'_');
      w3:=EatStr(s,'_');
      CreateInt;
     end else begin
      n:=AtomPtr-1;
      while (n>=0) and (Atoms[n]<>s) do dec(n);
      if n<0 then begin
       n:=AtomPtr; inc(AtomPtr);
       if AtomPtr>MaxStr then Error('Atom space overflow');
       Atoms[n]:=s;
      end;
      SetAtom(reg,atom,n);
     end;
   end;
 end;
 
procedure ReadSExp(r0,r1:byte; p:PStream);
 var s:string;
 function GetLex:string;
  var ok:boolean;
  begin 
   repeat 
    GetLex:=FGetLex(p,ok);
    if not ok then Delay(10);
   until ok 
  end;
 procedure ReadTail; forward; 
 procedure ReadSExp0;
  begin
   s:=GetLex;
   if s='(' then ReadTail else NewAtom(s,r0)
  end;
 procedure ReadTail; 
  procedure Tail0;
   begin
    Push(r0,r1); ReadTail; Roll(r1,r0);
   end;
  begin
   s:=GetLex;
   if s='('
    then begin ReadTail; Tail0 end
    else if s=')'
     then SetAtom(r0,null,0) 
     else if s='.' 
      then begin 
       ReadSExp0; s:=GetLex;
       if s<>')' then Error('Bad S-Expresion');
      end 
      else begin NewAtom(s,r0); Tail0 end;
  end;
 begin
  ReadSExp0;
 end;

procedure ReadSFrom(fname:string; r:byte);
 var p:PStream;
 begin
  p:=FOpen(fname,instream);
  ReadSExp(r,MaxReg,p); 
  FClose(p);
 end;
 
procedure FLink(fn:string);
 begin close(stderr); assign(stderr,fn); rewrite(stderr) end;

procedure WriteSTo(fname:string; r:byte);
 begin
  FLink(fname);
  ShowReg(r,99999,true); writeln(stderr);
  FLink('');
 end;
 
begin
 FLink('secd.log');
 AtomPtr:=0;
end.


