unit secd_mem;
interface

const
 MaxReg=10;
type
 TTag = (null,node,int,atom);
 TReg = record
  t:TTag;
  v:longint
 end;
 TNode = record
  c:longint; { Reference counter }
  l,r:TReg;
 end;

var
 Regs: array[0..MaxReg] of TReg;
 NodePtr:longint;

procedure SetAtom(reg:byte; t:TTag; v:longint);
procedure ShowReg(reg:byte; limit:longint; lnCase:boolean);
procedure ShowStat;
procedure Error(s:string);

procedure Car(s,d:byte);
procedure CarSelf(s:byte);
procedure Cdr(s,d:byte);
procedure CdrSelf(s:byte);
procedure GetNth(s,d:byte; i:word);
procedure Pop(s,d:byte);
procedure Cons(l,r,d:byte);
procedure Push (s,d:byte);
procedure PushR(s,d:byte);
procedure Roll(s,d:byte);
procedure MoveReg(s,d:byte);
procedure SetCar(s,d:byte);

implementation

uses secd_str, secd_cod;

const
 MaxNode = 10000;
var
 Nodes:array[0..MaxNode] of TNode;
 ConsCnt:longint;

{---------- Reference Count Garbage Collector ------------}

const
 RCMax=100;
var 
 RCAvail:TReg;
 RCAvCnt:longint;
 RCStack:array[0..RCMax] of longint;
 RCStPtr:word;
 
procedure RC_Collect(v1:longint);
 var w:TReg; v:longint; finish:boolean;
 begin
  v:=v1;
  repeat
  
   { Delay release of right branch of the tree }
   w:=Nodes[v].r;
   if w.t=node then begin
    RCStack[RCStPtr]:=w.v;
    if RCStPtr<=RCMax 
     then inc(RCStPtr)
     else Error('Garbage collector stack overflow');
   end;
   
   { Immediate release of left branch of the tree }
   w:=Nodes[v].l;
   
   Nodes[v].l:=RCAvail;
   RCAvail.t:=node;
   RCAvail.v:=v;
   inc(RCAvCnt);

   finish:=true;
   if w.t=node then begin
    v:=w.v;
    dec(Nodes[v].c);
    finish:=Nodes[v].c>0;
   end;
   
  until finish; 
 end;

procedure RC_inc(w:TReg);
 begin
  if w.t=node then inc(Nodes[w.v].c)
 end;

procedure RC_dec0(v:longint);
 begin
  dec(Nodes[v].c);
  if Nodes[v].c=0 then RC_Collect(v);
  if Nodes[v].c<0 then Error('RefCount<0');
 end;

procedure RC_dec(w:TReg);
 begin
  if w.t=node then RC_dec0(w.v)
 end;

{---------- List-oriented meta_commands ------------------}

procedure Car(s,d:byte);
 var w:TReg;
 begin 
  if s<>d then begin
   w:=Regs[s];
   if w.t=node then begin
    RC_dec(Regs[d]);
    Regs[d]:=Nodes[w.v].l;
    RC_inc(Regs[d]);
   end else Error('Car of an atom')
  end else Error('Use CdrSelf(x)')
 end;
 
procedure Cdr(s,d:byte);
 var w:TReg;
 begin 
  if s<>d then begin
   w:=Regs[s];
   if w.t=node then begin
    RC_dec(Regs[d]);
    Regs[d]:=Nodes[w.v].r;
    RC_inc(Regs[d]);
   end else Error('Cdr of an atom')
  end else Error('Use CdrSelf(x)')
 end;

procedure CarSelf(s:byte);
 var w:TReg;
 begin 
  w:=Regs[s];
  if w.t=node then begin
   Regs[s]:=Nodes[w.v].l;
   RC_inc(Regs[s]);
   RC_dec(w);
  end else Error('CarSelf of an atom')
 end;
 
procedure CdrSelf(s:byte);
 var w:TReg;
 begin 
  w:=Regs[s];
  if w.t=node then begin
   Regs[s]:=Nodes[w.v].r;
   RC_inc(Regs[s]);
   RC_dec(w);
  end else Error('CdrSelf of an atom')
 end;

procedure GetNth(s,d:byte; i:word);
 var j:word; w:TReg;
 begin
  if s<>d then begin
   j:=i;
   w:=Regs[s];
   while j>0 do begin
    if w.t=node 
     then w:=Nodes[w.v].r
     else Error('List for GetNth too short');
    dec(j); 
   end;
   RC_dec(Regs[d]);
   if w.t=node
    then Regs[d]:=Nodes[w.v].l
    else Error('i-th element is atom in GetNth');
   RC_inc(Regs[d]);
  end else Error('GetNth(s,i,s) is impossible');
 end;

procedure Pop(s,d:byte);
 var w:TReg;
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   RC_dec(Regs[d]);
   Regs[d]:=Nodes[w.v].l;
   RC_inc(Regs[d]);
   Regs[s]:=Nodes[w.v].r;
   RC_inc(Regs[s]);
   RC_dec(w);
  end else Error('Pop of an atom')
 end else Error('Pop(x,x) is impossible')
 end;
 
procedure GetNewNode(var v:longint);
 begin
 
  { Lazy Reference Count Garbage Collector step }
  if RCStPtr>0 then begin
   dec(RCStPtr);
   RC_dec0(RCStack[RCStPtr]);
  end;
  
  { Force RefCount Collector if RCAvail=nil }
  if RCAvail.t<>node then while RCStPtr>0 do begin
   dec(RCStPtr);
   RC_dec0(RCStack[RCStPtr]);
  end;
  
  if RCAvail.t=node then begin
   dec(RCAvCnt);
   v:=RCAvail.v;
   RCAvail:=Nodes[v].l;
  end else begin
   v:=NodePtr; inc(NodePtr);
   if NodePtr>MaxNode then Error('Heap overflow');
  end;
  inc(ConsCnt);
 end;
 
procedure Cons(l,r,d:byte);
 var v:longint;
 begin
  if (l<>d) and (r<>d) then begin
   GetNewNode(v);
   Nodes[v].l:=Regs[l];
   Nodes[v].r:=Regs[r];
   RC_inc(Regs[l]);
   RC_inc(Regs[r]);
   RC_dec(Regs[d]);
   Nodes[v].c:=1;
   Regs[d].t:=node;
   Regs[d].v:=v;
  end else Error('Use Push or PushR'); 
 end;

procedure Push (s,d:byte);
 var v:longint;
 begin
  if s<>d then begin
   GetNewNode(v);
   Nodes[v].l:=Regs[s];
   Nodes[v].r:=Regs[d];
   RC_inc(Regs[s]);
   Nodes[v].c:=1;
   Regs[d].t:=node;
   Regs[d].v:=v;
  end else Error('Push(x,x) is impossible'); 
 end;
 
procedure PushR(s,d:byte);
 var v:longint;
 begin
  if s<>d then begin
   GetNewNode(v);
   Nodes[v].l:=Regs[d];
   Nodes[v].r:=Regs[s];
   RC_inc(Regs[s]);
   Nodes[v].c:=1;
   Regs[d].t:=node;
   Regs[d].v:=v;
  end else Error('PushR(x,x) is impossible'); 
 end;
 
procedure Roll(s,d:byte);
 var w1,w2:TReg; v:longint;
 begin if s<>d then begin
  w1:=Regs[s];
  if w1.t=node then begin
   Regs[s]:=Nodes[w1.v].r;
   w2:=     Nodes[w1.v].l;
   RC_inc(Regs[s]);
   RC_inc(w2);
   RC_dec(w1);
   GetNewNode(v);
   Nodes[v].l:=w2;
   Nodes[v].r:=Regs[d];
   Nodes[v].c:=1;
   Regs[d].t:=node;
   Regs[d].v:=v;
  end else Error('Roll of an atom')
 end else Error('Roll(x,x) is impossible') 
 end;
 
procedure MoveReg(s,d:byte);
 begin 
  if s<>d then begin
   RC_dec(Regs[d]);
   Regs[d]:=Regs[s];
   RC_inc(Regs[d]);
  end else Error('MoveReg(s,s) is impossible');  
 end;
 
procedure SetCar(s,d:byte);
 var w:TReg;
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   if Nodes[w.v].l.t<>null then Error('SetCar allowed only for (NIL.x) nodes');
   Nodes[w.v].l:=Regs[d];
   RC_inc(Regs[d]);
  end else Error('SetCar to an atom')
 end else Error('SetCar(x,x) is impossible')
 end;

procedure SetAtom(reg:byte; t:TTag; v:longint);
 begin if (reg<=MaxReg) and (t<>node) then begin
  RC_dec(Regs[reg]);
  Regs[reg].t:=t; Regs[reg].v:=v;
 end end;

{------------------------ Misc ---------------------------}

procedure Error(s:string);
 begin
  writeln;
  writeln(s,' !');
  Halt;
 end;
procedure ShowStat;
 begin
  write('--- SECD : '); 
  write  ('ConsUse=',ConsCnt:6,' ');
  write  ('  SECD_loops=',LoopCnt:6,' ');
  writeln('  AtomUse=',AtomPtr,' ');
  write('--- stats: '); 
  write  ('HeapUse=',NodePtr:6,' ');
  write  ('  FreeNodes= ',RCAvCnt:6,' ');
  writeln('  GCStPtr=',RCStPtr,' ');
 end;

{---------------------------------------------------------}

procedure ShowReg(reg:byte; limit:longint; lnCase:boolean);
 var cnt,lcnt:longint; depth,d0:word;
 procedure wr(s:string);
  var i:byte;
  begin
   for i:=1 to length(s) do begin
    inc(cnt); inc(lcnt);
    if cnt<limit then write(fout,s[i]);
   end;
  end;
 procedure ShowReg_0(R:TReg); forward;
 procedure ShowTail(R:TReg);  forward;
 procedure ShowAtom(R:TReg);
  var s:string;
  begin
   case R.t of
    int:Str(R.v,s);
    atom:s:=GetStr(R.v);
   end;
   wr(s);
  end;
 procedure ShowNode(v:longint);
  var w:TReg;
  begin
   w:=Nodes[v].l;
   ShowReg_0(w);
   w:=Nodes[v].r;
   ShowTail(w);
  end;
 procedure ShowTail(R:TReg);
  begin if cnt<limit then
   case R.t of
    node:begin wr(' '); ShowNode(R.v) end;
    int,atom:begin wr('.'); ShowAtom(R) end;
    null:;
   end; 
  end;
 procedure ShowReg_0(R:TReg);
  var i:word;
  begin if cnt<limit then 
   case R.t of
    null:wr('NIL');
    node:begin 
          if (lcnt-3*(depth-d0)>40) and lnCase then begin
           writeln(fout); lcnt:=0; d0:=depth;
           for i:=0 to depth do wr(' ');
          end;
          wr('('); inc(depth); ShowNode(R.v); dec(depth); wr(')');
         end;
    int,atom:ShowAtom(R);
   end;
  end;
 begin
  cnt:=0; lcnt:=0; depth:=0; d0:=0;
  if (reg<=MaxReg) then ShowReg_0(Regs[reg]) 
 end;
 
procedure InitMem;
 var i:byte;
 begin
  for i:=0 to MaxReg do Regs[i].t:=null;
  NodePtr:=0; ConsCnt:=0;
  RCAvCnt:=0; RCAvail.t:=null; RCStPtr:=0;
 end;

begin
 InitMem;
end.

