unit secd_mem;
interface

const
 MaxReg=10;
type
 TTag = (null,node,int,atom);
 TReg = record
  t:TTag;
  v:longint
 end;
 TNode = record
  c:longint; { Reference counter }
  l,r:TReg;
 end;

var
 Regs: array[0..MaxReg] of TReg;
 NodePtr:longint;

procedure SetAtom(reg:byte; t:TTag; v:longint);
procedure ShowReg(reg:byte; limit:longint; lnCase:boolean);
procedure ShowStat;
procedure Error(s:string);

procedure Car(s,d:byte);
procedure Cdr(s,d:byte);
procedure Pop(s,d:byte);
procedure Cons(l,r,d:byte);
procedure Roll(s,d:byte);
procedure MoveReg(s,d:byte);
procedure SetCar(s,d:byte);

implementation

uses secd_str, secd_cod;

const
 MaxNode = 99000;
var
 Nodes:array[0..MaxNode] of TNode;

procedure Error(s:string);
 begin
  writeln;
  writeln(s,' !');
  Halt;
 end;
procedure ShowStat;
 begin
  write('--- SECD stats: '); 
  write('Cons=',NodePtr,' ');
  write('Atoms=',AtomPtr,' ');
  write('Loops=',LoopCnt,' ');
  writeln;
 end;
 
procedure Car(s,d:byte);
 var w:TReg;
 begin if s<=MaxReg then begin
  w:=Regs[s];
  if w.t=node then begin
   Regs[d]:=Nodes[w.v].l;
  end else Error('Car of an atom')
 end end;
 
procedure Cdr(s,d:byte);
 var w:TReg;
 begin if s<=MaxReg then begin
  w:=Regs[s];
  if w.t=node then begin
   Regs[d]:=Nodes[w.v].r;
  end else Error('Cdr of an atom')
 end end;
 
procedure Pop(s,d:byte);
 var w:TReg;
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   Regs[d]:=Nodes[w.v].l;
   Regs[s]:=Nodes[w.v].r;
  end else Error('Pop of an atom')
 end else Error('Pop(x,x) is inpossible')
 end;
 
procedure GetNewNode(var v:longint);
 begin
  v:=NodePtr; inc(NodePtr);
  if NodePtr>MaxNode then Error('Heap overflow');
 end;
 
procedure Cons(l,r,d:byte);
 var v:longint;
 begin
  GetNewNode(v);
  Nodes[v].l:=Regs[l];
  Nodes[v].r:=Regs[r];
  Nodes[v].c:=1;
  Regs[d].t:=node;
  Regs[d].v:=v;
 end;
 
procedure Roll(s,d:byte);
 var w:TReg; v:longint;
 begin if s<>d then begin
  w:=Regs[s];
  if w.t=node then begin
   Regs[s]:=Nodes[w.v].r;
   w:=Nodes[w.v].l;
   GetNewNode(v);
   Nodes[v].l:=w;
   Nodes[v].r:=Regs[d];
   Nodes[v].c:=1;
   Regs[d].t:=node;
   Regs[d].v:=v;
  end else Error('Roll of an atom')
 end else Error('Roll(x,x) is inpossible') 
 end;
 
procedure MoveReg(s,d:byte);
 begin Regs[d]:=Regs[s] end;
procedure SetCar(s,d:byte);
 var w:TReg;
 begin if s<=MaxReg then begin
  w:=Regs[s];
  if w.t=node then begin
   Nodes[w.v].l:=Regs[d];
  end else Error('SetCar to an atom')
 end end;

procedure SetAtom(reg:byte; t:TTag; v:longint);
 begin if (reg<=MaxReg) and (t<>node) then begin
  Regs[reg].t:=t; Regs[reg].v:=v;
 end end;

procedure ShowReg(reg:byte; limit:longint; lnCase:boolean);
 var cnt,lcnt:longint; depth,d0:word;
 procedure wr(s:string);
  var i:byte;
  begin
   for i:=1 to length(s) do begin
    inc(cnt); inc(lcnt);
    if cnt<limit then write(fout,s[i]);
   end;
  end;
 procedure ShowReg_0(R:TReg); forward;
 procedure ShowTail(R:TReg);  forward;
 procedure ShowAtom(R:TReg);
  var s:string;
  begin
   case R.t of
    int:Str(R.v,s);
    atom:s:=GetStr(R.v);
   end;
   wr(s);
  end;
 procedure ShowNode(v:longint);
  var w:TReg;
  begin
   w:=Nodes[v].l;
   ShowReg_0(w);
   w:=Nodes[v].r;
   ShowTail(w);
  end;
 procedure ShowTail(R:TReg);
  begin if cnt<limit then
   case R.t of
    node:begin wr(' '); ShowNode(R.v) end;
    int,atom:begin wr('.'); ShowAtom(R) end;
    null:;
   end; 
  end;
 procedure ShowReg_0(R:TReg);
  var i:word;
  begin if cnt<limit then 
   case R.t of
    null:wr('NIL');
    node:begin 
          if (lcnt-3*(depth-d0)>40) and lnCase then begin
           writeln(fout); lcnt:=0; d0:=depth;
           for i:=0 to depth do wr(' ');
          end;
          wr('('); inc(depth); ShowNode(R.v); dec(depth); wr(')');
         end;
    int,atom:ShowAtom(R);
   end;
  end;
 begin
  cnt:=0; lcnt:=0; depth:=0; d0:=0;
  if (reg<=MaxReg) then ShowReg_0(Regs[reg]) 
 end;
 
procedure InitMem;
 var i:byte;
 begin
  for i:=0 to MaxReg do Regs[i].t:=null;
  NodePtr:=0;
 end;

begin
 InitMem;
end.

