unit secd_tcp;

interface

uses Sockets, Linux;

type
 PTCPSocket = ^TTCPSocket;
 TCPConnTyp  = (listener, conn_from, conn_to);
 TTCPSocket = record
  conn:longint;
  SA:TInetSockAddr;
  t:TCPConnTyp;
 end;

function TCPListen (aPort:word; maxconn:word):PTCPSocket;
function TCPConnect(aPort:word; host:longint):PTCPSocket;
function TCPGetConn(socket:PTCPSocket; var ok:boolean):PTCPSocket;

implementation

uses secd_mem;

const
 Err_TryAgain=11;
 Err_InProgress=115;
 
function SwapWord(w:word):word;
 begin
  SwapWord:=(w shl 8) or (w shr 8);
 end;
    
function TCPSocket(aPort:word):PTCPSocket;
 var p:PTCPSocket;
 begin
  New(p);
  with p^ do begin
   conn:=Socket(AF_INET,SOCK_STREAM,0 {6} {PF_INET});
   if conn<0 then Error('socket');
   fcntl(conn,F_SetFl,Open_NonBlock);
   with SA do begin
    family:=AF_INET;
    port:=SwapWord(aPort);
   end; 
  end;  
  TCPSocket:=p;
 end;
 
function TCPListen (aPort:word; maxconn:word):PTCPSocket;
 var p:PTCPSocket;
 begin
  p:=TCPSocket(aPort);
  with p^ do begin
   SA.addr:=0; { allow connections from anywhere }
   t:=listener;
   if not Bind(conn,SA,SizeOf(SA)) then Error('bind');
   if not Listen(conn,maxconn)     then Error('listen');
  end;  
  TCPListen:=p;
 end;
 
function TCPConnect(aPort:word; host:longint):PTCPSocket;
 var p:PTCPSocket;
 begin
  p:=TCPSocket(aPort);
  with p^ do begin
   SA.addr:=host; { ((1 shl 24) or 127) for localhost } 
   t:=conn_to;
  end;  
  TCPConnect:=p;
 end;
 
function TCPGetConn(socket:PTCPSocket; var ok:boolean):PTCPSocket;
 var p:PTCPSocket; addrlen:longint;
 begin 
  case socket^.t of
   listener:begin
     New(p);
     p^.t:=conn_from;
     p^.conn:=Accept(socket^.conn,p^.SA,addrlen);
     ok:=p^.conn>=0;
     if ok 
      then TCPGetConn:=p
      else if SocketError<>Err_TryAgain 
       then Error('accept')
       else Dispose(p);     
    end;
   conn_to:begin
     with socket^ do ok:=Connect(conn,SA,SizeOf(SA));
     if (not ok) and (SocketError<>Err_InProgress) then begin
      writeln(stderr,'SocketError=',SocketError);
      Error('connect');
     end;
     TCPGetConn:=socket;
    end;
  end;  
 end;

end.

