unit secd_msg;

interface

uses secd_mem;

procedure SendMsg(mode:byte);
procedure RecvMsg(mode:byte);
procedure ProcLoad;
procedure ProcSave;
procedure IncPend;
procedure IO_Scan;

{

 Pipe is passive - (q0.q1) or waiting - [pipe: s' (e'.c')]

 -------------------
 mode=0

 SEND
 (Pipe   msg s) e c --> ((msg s) e c)
 (Stream msg s) e c --> ((T   s) e c)

 RECV
 (Pipe   s) e c   --> ((msg_from(Pipe) s)   e c)
 (Stream s) e c   --> ((getlex(Stream) s)   e c) 
 (Connwait s) e c --> ((socket(Connwait) s) e c)

 -------------------
 mode=1
 
 RECVC (Stream s) e c     --> Receive char (ascii code) from Stream
 SENDC (Stream I s) e c   --> Send char with ascii code I to Stream
  

}

implementation

uses secd_cod, secd_str, secd_tcp;

procedure ProcLoad;
 begin
  if Regs[r1].t=null then while Regs[r0].t=node do Roll(r0,r1);
  while (Regs[r1].t=null) and (Regs[io].t=node) do begin
   IO_Scan;
   if Regs[r1].t=null then Delay;
  end;
  if Regs[r1].t=node then begin
   Pop(r1,c);
   Pop(c,s);
   Pop(c,e);
   ChCnt:=0;
   SwapOK:=false;
  end;
  if debug then begin
   writeln(stderr,'---- New Proc ---- ProcCnt=',PendingCnt);
   ShowSECD;
  end; 
  CPCnt:=0;
  dec(PendingCnt); 
 end;

procedure IncPend;
 begin
  inc(PendingCnt);
  if PendingCnt>PendMax then PendMax:=PendingCnt;
 end;
 
procedure ProcSave;
 begin
  Push(e,c);
  Push(s,c);
  Push(c,r0); 
  IncPend;
 end;

procedure IOSave;
 begin
  Push(e,c);
  Push(s,c);
  SetCdr(a,c);
  Push(a,io); 
  SwapOK:=true;
 end;

procedure SendMsg(mode:byte);
 var ws:string; ok:boolean;
 begin
  Just(1);
  if (mode>0) and (SubTag(a)<>stream)
   then Error('Impossible sendX');
  case SubTag(a) of
   cons,pipe:
    ProcSave; { Delay current process: r0 --> (((msg s) e c) r0) }
  end;  
  ChCnt:=0;
  Pop(s,b); { a=Pipe/Stream, b:=msg } 
  case SubTag(a) of
   cons:begin
      Car(a,w1);
      Push(b,w1);
      SetCar(a,w1);
      SwapOK:=true;
     end;
   pipe:begin
      Car(a,s);
      Cdr(a,c); Pop(c,e);
      SetAtom(w1,null,0);
      SetCar(a,w1); SetCdr(a,w1);
      SetSubTag(a,cons);
      CPCnt:=0;
      MoveReg(b,a);
      ChCnt:=1;
     end;
   stream:begin
      Car(a,w1);
      case mode of
       0:begin
          case Regs[b].t of
           null:ws:='nil';
           int:ws:=GetLong(Regs[b].v);
           atom:ws:=GetStr(Regs[b].v);
           node:Error('Sending node to stream is impossible');
          end; 
          ok:=FPutStr(PStream(Regs[w1].v),ws) 
         end;  
       1:begin
          if Regs[b].t<>int
           then Error('SendC allow only integers');
          ok:=FPutChar(PStream(Regs[w1].v),Chr(Regs[b].v)) 
         end;
      end;
      if ok 
       then begin NewAtom('T',a); ChCnt:=1 end 
       else IOSave;
     end;
  end; 
 end;

procedure CreateSocketNode(socket:PTCPSocket);
 var p:PStream;
 begin
  p:=SOpen(socket,instream);
  CreateIO(longint(p),stream);
  MoveReg(a,w1);
  p:=SOpen(socket,outstream);
  CreateIO(longint(p),stream);
  Push(w1,a);
 end;
 
procedure RecvMsg(mode:byte);
 var ok:boolean; ws:string; wc:char; socket:PTCPSocket;
 begin
  Just(1);
  ChCnt:=0;
  if (mode>0) and (SubTag(a)<>stream)
   then Error('Impossible recvX');
  case SubTag(a) of
   cons:begin
      Car(a,b); Cdr(a,w1);
      if Regs[w1].t=null then while Regs[b].t=node do Roll(b,w1);
      if Regs[w1].t=null then begin
       Push(e,c);
       SetCar(a,s); SetCdr(a,c);
       SetSubTag(a,pipe);
       SwapOK:=true;
      end else begin
       SetCar(a,b); 
       Pop(w1,b);
       SetCdr(a,w1);
       MoveReg(b,a);
       ChCnt:=1;
      end;
     end;
   pipe:Error('Somebody wait on this pipe');
   stream:begin
      Car(a,w1);
      case mode of
       0:begin
          ws:=FGetLex(PStream(Regs[w1].v),ok);
          if ok then NewAtom(ws,a); 
         end;
       1:begin
          wc:=FGetChar(PStream(Regs[w1].v),ok);
          if ok then SetAtom(a,int,ord(wc));  
         end;
      end;   
      if ok then ChCnt:=1 else IOSave;
     end;
   connwait:begin
      Car(a,w1);
      socket:=TCPGetConn(PTCPSocket(Regs[w1].v),ok);
      if ok then begin
       CreateSocketNode(socket);
       ChCnt:=1;
      end else IOSave;
     end;
  end; 
 end;

procedure IO_Scan;
 var 
  ok:boolean; 
  ws:string;
  wc:char;
  ps:PStream;
  socket:PTCPSocket;
 procedure IOLoad0;
  begin
   { NewAtom(ws,w1);} { w1 = GetLex or 'T' }
   Cdr(b,c);       { c  = (s' e' c') }
   Pop(c,s); 
   Push(w1,s);
   Push(s,c);      { c  = ((w1 s') e' c') }
   SetAtom(e,null,0);
   SetCdr(b,e);    { b= <stream/connwait|pFile.nil> } 
   Push(c,r1);     { process ready }
   IncPend;
  end;
 procedure IOLoad(ws:string);
  begin
   NewAtom(ws,w1); { w1 = GetLex or 'T' }
   IOLoad0;
  end; 
 procedure IOLoadC;
  begin
   SetAtom(w1,int,ord(wc)); { w1 = ord(GetChar) }
   IOLoad0;
  end; 
 begin
  SetAtom(a,null,0);
  while Regs[io].t=node do begin
   Pop(io,b); { b = <stream/connwait|pFile.(s' e' c')> }
   Car(b,w1); { w1 = pFile/pSocket }
   case SubTag(b) of
    stream:begin
      ps:=PStream(Regs[w1].v);
      case ps^.t of 
       instream:case ps^.srmode of
          0:begin
             ws:=FGetLex(ps,ok);
             if ok then IOLoad(ws);
            end;
          1:begin
             wc:=FGetChar(ps,ok);
             if ok then IOLoadC;
            end;
         end;
       outstream:begin
          case ps^.srmode of
           0:ok:=FPutStr(ps,ps^.lex);
           1:ok:=FPutChar(ps,ps^.w);
          end;
          if ok then IOLoad('T');
         end; 
      end;
     end;
    connwait:
      begin
       socket:=TCPGetConn(PTCPSocket(Regs[w1].v),ok);
       if ok then begin
        MoveReg(a,s); MoveReg(b,e);
        CreateSocketNode(socket);
        MoveReg(a,w1);
        MoveReg(s,a); MoveReg(e,b);
        IOLoad0;
       end;
      end;
    else Error('Bad IO object');  
   end;
   if not ok then Push(b,a);
  end;
  MoveReg(a,io);
 end;

end.

